# -*- coding: utf-8 -*-
"""
Created on Tue Jul 23 16:43:54 2024

@author: atakan
"""

import json
import yaml
import carbatpy as cb
res_dir =cb._RESULTS_DIR+'\\hp-input-dict'

# Zuweisung der Variablen und Parameter
FLUID = "Propane * Butane * Pentane * Hexane"
comp = [0.75, 0.05, 0.15, 0.05]
# comp = [0.4, 0.3, 0.3, 0.0]  # [0.164, 0.3330, 0.50300, 0.0]

FLS = "Water"
FLCOLD = "Methanol"
ETA_S_ = 0.7
_STORAGE_T_IN_ = 298.15  # Beispielwert für cb._T_SURROUNDING
_COLD_STORAGE_T_IN_ = _STORAGE_T_IN_
_STORAGE_T_OUT_ = 363.0  # 395.0
_COLD_STORAGE_T_OUT_ = 250.15
_STORAGE_P_IN_ = 5e5
_COLD_STORAGE_P_IN_ = 5e5
_Q_DOT_MIN_ = 1e3  # heat flow rate (W)
_D_T_SUPER_ = 5  # super heating of working fluid
_D_T_MIN_ = 4.0  # minimum approach temperature (pinch point)

# Speichern der Variablen und Parameter in einem Dictionary
variables_dict = {
    'FLUID': FLUID,
    'comp': comp,
    'FLS': FLS,
    'FLCOLD': FLCOLD,
    'ETA_S_': ETA_S_,
    '_STORAGE_T_IN_': _STORAGE_T_IN_,
    '_COLD_STORAGE_T_IN_': _COLD_STORAGE_T_IN_,
    '_STORAGE_T_OUT_': _STORAGE_T_OUT_,
    '_COLD_STORAGE_T_OUT_': _COLD_STORAGE_T_OUT_,
    '_STORAGE_P_IN_': _STORAGE_P_IN_,
    '_COLD_STORAGE_P_IN_': _COLD_STORAGE_P_IN_,
    '_Q_DOT_MIN_': _Q_DOT_MIN_,
    '_D_T_SUPER_': _D_T_SUPER_,
    '_D_T_MIN_': _D_T_MIN_,
}

# Funktion zum Wiederzuweisen der Variablen aus dem Dictionary
def assign_variables_from_dict(variables_dict):
    global FLUID, comp, FLS, FLCOLD, ETA_S_, _STORAGE_T_IN_
    global _COLD_STORAGE_T_IN_, _STORAGE_T_OUT_, _COLD_STORAGE_T_OUT_
    global _STORAGE_P_IN_, _COLD_STORAGE_P_IN_, _Q_DOT_MIN_, _D_T_SUPER_, _D_T_MIN_
    
    FLUID = variables_dict['FLUID']
    comp = variables_dict['comp']
    FLS = variables_dict['FLS']
    FLCOLD = variables_dict['FLCOLD']
    ETA_S_ = variables_dict['ETA_S_']
    _STORAGE_T_IN_ = variables_dict['_STORAGE_T_IN_']
    _COLD_STORAGE_T_IN_ = variables_dict['_COLD_STORAGE_T_IN_']
    _STORAGE_T_OUT_ = variables_dict['_STORAGE_T_OUT_']
    _COLD_STORAGE_T_OUT_ = variables_dict['_COLD_STORAGE_T_OUT_']
    _STORAGE_P_IN_ = variables_dict['_STORAGE_P_IN_']
    _COLD_STORAGE_P_IN_ = variables_dict['_COLD_STORAGE_P_IN_']
    _Q_DOT_MIN_ = variables_dict['_Q_DOT_MIN_']
    _D_T_SUPER_ = variables_dict['_D_T_SUPER_']
    _D_T_MIN_ = variables_dict['_D_T_MIN_']

# Speichern des Dictionaries in einer JSON-Datei
with open(res_dir+'variables.json', 'w') as json_file:
    json.dump(variables_dict, json_file, indent=4)

# Einlesen des Dictionaries aus einer JSON-Datei
with open(res_dir+'variables.json', 'r') as json_file:
    variables_dict = json.load(json_file)

# Variablen aus dem Dictionary zuweisen
assign_variables_from_dict(variables_dict)

# Ausgabe, um zu prüfen, ob die Variablen korrekt zugewiesen wurden
print("JSON:")
print(FLUID)
print(comp)
print(FLS)
print(FLCOLD)
print(ETA_S_)
print(_STORAGE_T_IN_)
print(_COLD_STORAGE_T_IN_)
print(_STORAGE_T_OUT_)
print(_COLD_STORAGE_T_OUT_)
print(_STORAGE_P_IN_)
print(_COLD_STORAGE_P_IN_)
print(_Q_DOT_MIN_)
print(_D_T_SUPER_)
print(_D_T_MIN_)

# Speichern des Dictionaries in einer YAML-Datei
with open(res_dir+'variables.yaml', 'w') as yaml_file:
    yaml.dump(variables_dict, yaml_file, default_flow_style=False)

# Einlesen des Dictionaries aus einer YAML-Datei
with open(res_dir+'variables.yaml', 'r') as yaml_file:
    variables_dict = yaml.safe_load(yaml_file)

# Variablen aus dem Dictionary zuweisen
assign_variables_from_dict(variables_dict)

# Ausgabe, um zu prüfen, ob die Variablen korrekt zugewiesen wurden
print("\nYAML:")
print(FLUID)
print(comp)
print(FLS)
print(FLCOLD)
print(ETA_S_)
print(_STORAGE_T_IN_)
print(_COLD_STORAGE_T_IN_)
print(_STORAGE_T_OUT_)
print(_COLD_STORAGE_T_OUT_)
print(_STORAGE_P_IN_)
print(_COLD_STORAGE_P_IN_)
print(_Q_DOT_MIN_)
print(_D_T_SUPER_)
print(_D_T_MIN_)




# -*- coding: utf-8 -*-
"""
Created on Thu Jul 25 16:42:44 2024
heat-exchanger class

@author: welp
"""

import carbatpy as cb
import numpy as np
from scipy.optimize import fsolve
from scipy.integrate import solve_ivp
import matplotlib.pyplot as plt
import carbatpy.models.components.heat_transfer as heat_transfer
import time


class heat_exchanger:
    """
    class for static heat-exchanger, different configurations in future (?),
    start with double pipe counterflow heat-exchanger
    start: get outlet temperature with area A (fixed)
    
    """
    def __init__(self, fluids, inlet_states, mdot, resolution):
        self.fluids = fluids
        self.state_in_wf = inlet_states[0]
        self.state_in_sf = inlet_states[1]
        self.mdot_wf = mdot[0]
        self.mdot_sf = mdot[1]
        self.resolution = resolution
        
    def geo_A(self, A):
        self.A = A # please use Ai (inner tube surface) as reference
        self.include_pressure_loss = False
        
    def geo_double_pipe(self, di, Di, da, Da, l, lam_tube, method, incl_pr_loss=True):
        self.di = di
        self.Di = Di
        self.da = da
        self.Da = Da
        self.l = l
        self.A = np.pi * self.di * self.l
        self.method = method
        self.lam_tube= lam_tube
        self.include_pressure_loss = incl_pr_loss
        
    def calc_NTU(self, U_i=0):
        """
        calculate heat-exchanger with NTU cell-method

        Parameters
        ----------
        U_i : TYPE, optional
            DESCRIPTION. The default is 0. -> local heat-transfer coefficient
            is calculated. Then please define NTU-method. If set to value,
            constant U_i is assumed.

        Returns
        -------
        None.

        """
        
        T_guess = - (self.state_in_wf[0] - self.state_in_sf[0]) * 0.5 + \
                    self.state_in_wf[0]  # guesses start temperature for solver
        p_guess = self.state_in_sf[1]  # don't change, otherwise calculation without
                                        # pressure loss does not converge
        T_correkt, p_correct = fsolve(self._solve_compose_NTU, 
                                      [T_guess, p_guess], args=(U_i), 
                                      xtol = 0.0001)
        state1, state2 = self._compose_NTU([T_correkt, p_correct], U_i)
        self.state_out_wf = state1[-1]
        self.state_out_sf = state2[0]
        self.state_array_wf = state1
        self.state_array_sf = state2
        
    def calc_IVP(self, U_i=0):
        """
        calculates heat-exchanger with solving ODE using solve ivp

        Parameters
        ----------
        U_i : TYPE, optional
            DESCRIPTION. The default is 0. -> local heat-transfer coefficient
            is calculated. Then please define NTU-method. If set to value,
            constant U_i is assumed.

        Returns
        -------
        None.

        """
        
        T_guess = - (self.state_in_wf[0] - self.state_in_sf[0]) * 0.5 + \
                    self.state_in_wf[0]  # guesses start temperature for solver
        p_guess = self.state_in_sf[1] # don't change, otherwise calculation without
                                        # pressure loss does not converge
        T_correct, p_correct = fsolve(self._solve_ODE_ivp, [T_guess, p_guess], 
                                      args=(U_i), xtol = 0.0001)
        
        state_correct_sf = self.fluids[1].set_state([T_correct, p_correct], "TP")
        y0 = [self.state_in_wf[2],
              state_correct_sf[2],
              self.state_in_wf[1],
              state_correct_sf[1]]
        x = np.linspace(0, self.l, self.resolution)
        res = solve_ivp(lambda x, y: self._ODE_ivp(x, y, U_i), (x[0], x[-1]), y0, t_eval=x, rtol=0.1)
        
        state1 = []
        state2 = []
        
        for i in range(len(res.t)):
            state1.append(self.fluids[0].set_state([res.y[0][i], res.y[2][i]], "HP"))
            state2.append(self.fluids[1].set_state([res.y[1][i], res.y[3][i]], "HP", 
                                                   cb.fprop._TRANS_STRING))
        
        state1 = np.array(state1)
        state2 = np.array(state2)
        self.state_out_wf = state1[-1]
        self.state_out_sf = state2[0]
        self.state_array_wf = state1
        self.state_array_sf = state2
        
    def _solve_ODE_ivp(self, y_tp, U_i):
        """
        internal helper function, used to iterate solve ivp
        """
        
        T_guess, p_guess = y_tp
        x = np.linspace(0, self.l, self.resolution)
        state_guess_sf = self.fluids[1].set_state([T_guess, p_guess], "TP")
        y0 = [self.state_in_wf[2],
              state_guess_sf[2],
              self.state_in_wf[1],
              state_guess_sf[1]]
        res = solve_ivp(lambda x, y: self._ODE_ivp(x, y, U_i), (x[0], x[-1]), 
                        y0, t_eval=x, rtol=0.1)
        state2 = self.fluids[1].set_state([res.y[1][-1], res.y[3][-1]], "HP")
        
        return [self.state_in_sf[0] - state2[0],
                self.state_in_sf[1] - state2[1]]
    
        
    def _ODE_ivp(self, x, y, U_i=0):
        """ 
        internal function, ODE to call in solve_ivp
        """
        h_wf, h_sf, p_wf, p_sf = y
        
        state_wf_x = (self.fluids[0].set_state([h_wf, p_wf], "HP"))
        state_sf_x = (self.fluids[1].set_state([h_sf, p_sf], "HP", 
                                               cb.fprop._TRANS_STRING))
        
        if U_i == 0:
            R_ges, dp_wf, dp_sf= self.thermal_resistance(state_wf_x, state_sf_x)
            dp_wf = -dp_wf
        else:
            R_ges = 1/U_i
            dp_wf = 0 
            dp_sf = 0 
  
        delta_T = state_sf_x[0] - state_wf_x[0]   

        dh_wf = self.di * np.pi/R_ges * delta_T / self.mdot_wf
        dh_sf = self.di * np.pi/R_ges * delta_T / self.mdot_sf
        
        return np.array([dh_wf, dh_sf, dp_wf, dp_sf])

        
    def _get_theta(self, U_i, A_i, W_i, C_i, characteristic='counterflow'):
        """
        calculates theta = f(NTU, C) to use in NTU-method for fluid i, second 
        fluid j
    
        Parameters
        ----------
        U_i : TYPE
            DESCRIPTION. heat transfer coefficient W/(m^2K)
        A_i : TYPE
            DESCRIPTION. heat transfer area m^2
        W_i : TYPE
            DESCRIPTION. heat capacity flow W/K
        C_i : TYPE
            DESCRIPTION. W_i / W_j
        characteristic : TYPE, optional
            DESCRIPTION. The default is 'counterflow'.
    
        Returns
        -------
        theta : TYPE
            DESCRIPTION. dimensionless temperature difference (Ti_in - Ti_out) 
                                                            / (Ti_in - Tj_in)
    
        """
        NTU = U_i * A_i / W_i
        if characteristic == "counterflow":
            theta = (1 - np.exp(NTU * (C_i - 1))) \
                     / (1 - C_i * np.exp(NTU * (C_i - 1)))   
                     # VDI heat atlas (2013) C1-table 3                           
        elif characteristic == "parallelflow":
            theta = (1 - np.exp(- NTU * (C_i + 1))) \
                     / (1 + C_i) # VDI heat atlas (2013) C1-table 3
        return theta
    
    
    def _solve_compose_NTU(self, y, U_i):
        """
        helper function to solve compose_NTU
    
        Parameters
        ----------
        y : TYPE
            DESCRIPTION. [T_sf_outlet, p_sf_outlet] to guess
        Returns
        -------
        TYPE
            DESCRIPTION. difference prescribed inlet j temperature - 
            calculated inlet j temperature, and same for pressure 
    
        """    
        state1, state2 = self._compose_NTU(y, U_i)
        #print(state2[-1])
        return [self.state_in_sf[0] - state2[-1][0],
                self.state_in_sf[1] - state2[-1][1]]
    
    def _compose_NTU(self, y, U_i):
        """
        uses cell method NTU from VDI heat atlas. currently only for 
        counterflow due to energy balance. Starts with inlet states for fluids 
        i and j and alterates outlet temperature for fluid j. Entire heat 
        exchanger is subsequently calculated from left to right. Cell on the 
        left: A, cell on the right: Z. Can be called with solve_compose_NTU 
        or independantly.
    
        Parameters
        ----------
        y : TYPE
            DESCRIPTION. outlet temperature of secondary fluid in cell A [K]
                        and outlet pressure of secondary fluid in cell A [Pa]
        U_i : TYPE
            DESCRIPTION. The default is 0. -> local heat-transfer coefficient
            is calculated. Then please define NTU-method. If set to value,
            constant U_i is assumed.
    
        Returns
        -------
        state1 : TYPE
            DESCRIPTION. array of state arrays fluid i. Line 0 cell A, line 
            (-1) cell Z
        state2 : TYPE
            DESCRIPTION. array of state arrays fluid j. Line 0 cell A, line 
            (-1) cell Z
    
        """
        T2A, p2A = y
        
        A_array = np.full(self.resolution, self.A/self.resolution) 
        # each cell is calculated with same area
        
        # initialize help variables to iteratate
        T2 = np.zeros(len(A_array))
        W1 = np.zeros(len(A_array))
        W2 = np.zeros(len(A_array))
        C2 = np.zeros(len(A_array))
        state1 = np.zeros((len(A_array), len(self.state_in_wf)))
        state2 = np.zeros((len(A_array), len(self.state_in_sf)))
        
        # set start values in array
        state1[0] = self.state_in_wf
        T2[0] = T2A
        state2[0] = self.fluids[1].set_state([T2A, p2A], \
                                             "TP", cb.fprop._TRANS_STRING)
        
        # iterate over cells
        for i in range(len(A_array)-1):
            W2[i] = self.mdot_sf * state2[i][10]
            if 0 < state1[i][5] < 1: # heat capacity of working fluid infinite
                C2[i] = 0
            else: # fluid i also 1phase
                W1[i] = self.mdot_wf * self.fluids[0].set_state([state1[i][0], state1[i][1]]\
                                         , "TP", cb.fprop._TRANS_STRING)[10]
                C2[i] = W2[i] / W1[i]
            
            if U_i == 0: 
                R, dp_wf, dp_sf = self.thermal_resistance(state1[i], state2[i])
                U_local = 1/R
            else:
                U_local = U_i
            
            theta = self._get_theta(U_local, A_array[i], W2[i], C2[i])
            
            try:
                T2[i+1] = (theta * state1[i][0] - T2[i]) / (theta-1)
            except:
                print("wait")
            if self.include_pressure_loss == True:
                delta_p_wf = -dp_wf * self.l / self.resolution
                delta_p_sf = dp_sf * self.l / self.resolution
                
            else:
                delta_p_wf = 0
                delta_p_sf = 0
            
            state2[i+1] = self.fluids[1].set_state([T2[i+1], 
                                                    state2[i][1] + delta_p_sf], 
                                                   "TP", 
                                                   cb.fprop._TRANS_STRING
                                                   )
            h1B = self.mdot_sf / self.mdot_wf \
                * (state2[i+1][2] - state2[i][2]) + state1[i][2]
            try:
                state1[i+1] = self.fluids[0].set_state([h1B, state1[i][1] + delta_p_wf], 
                                                   "HP")
            except:
                print("wait")
        
        return state1, state2
    
    def thermal_resistance(self, state_wf_x, state_sf_x):
        """
        thermal resistance for double-pipe

        Parameters
        ----------
        state_wf_x : TYPE
            DESCRIPTION.
        state_sf_x : TYPE
            DESCRIPTION.

        Returns
        -------
        R_ges : TYPE
            DESCRIPTION.

        """
        fluidstate_wf_x = self.fluids[0].set_state([state_wf_x[0], 
                                                    state_wf_x[1]], "TP", 
                                                   output="FluidState")
        fluidstate_sf_x = self.fluids[1].set_state([state_sf_x[0], 
                                                    state_sf_x[1]], "TP", 
                                                   cb.fprop._TRANS_STRING, 
                                                   output="FluidState")
        
        alpha_sf, dp_sf = heat_transfer.alpha_1P(self.fluids[1], fluidstate_sf_x, self.mdot_sf, 
                                                 self.l, self.da, self.Di)
        alpha_wf, dp_wf = heat_transfer.alpha_km(self.fluids[0], fluidstate_wf_x, 
                                                 self.mdot_wf, self.method, 
                                                 self.l, self.di)
        
        # heat transfer resistance
        R_sf = self.di/(alpha_sf*self.Di) # convection secondary fluid
        R_tube = np.log(self.Di/self.di) * self.di / self.lam_tube # conduction tube
        R_wf = 1/alpha_wf # convection working fluid

        R_ges = R_sf + R_tube + R_wf
        return R_ges, dp_wf, dp_sf
    
    def diagram(self, ordinate=0, second_yaxis=False):
        """
        plots diagram, abscissa area, ordinate can be chosen in same order of 
        state array: ("Temperature", "Pressure", "spec_Enthalpy",
                        "spec_Volume", "spec_Entropy", "quality",
                        "spec_internal_Energy",
                        "viscosity", "thermal_conductivity",
                        "Prandtl_number", "k_viscosity", "molecular_mass",
                        "speed_of_sound")

        Parameters
        ----------
        ordinate : TYPE, optional
            DESCRIPTION. The default is 0.

        Returns
        -------
        None.

        """        

        fig, ax1 = plt.subplots()
        x = np.linspace(0, self.resolution-1, self.resolution) \
                     * self.A/self.resolution

        ax1.plot(x, self.state_array_wf[:, ordinate], 'r:o', 
                 label='Working Fluid', markersize=5)
        ax1.set_xlabel('A [m^2]')
        
        if second_yaxis == True:
            ax1.tick_params(axis='y', labelcolor='r')
            ax1.set_ylabel(cb.CB_DEFAULTS["Fluid_Defaults"]['Property_Names'][ordinate], 
                           color='r')

            ax2 = ax1.twinx()
            # Plot the secondary fluid on the second y-axis
            ax2.plot(x, self.state_array_sf[:, ordinate], 'b--s', 
                     label='Secondary Fluid', markersize=5)
            ax2.set_ylabel(cb.CB_DEFAULTS["Fluid_Defaults"]['Property_Names'][ordinate], 
                           color='b')
            ax2.tick_params(axis='y', labelcolor='b')
        else:
            ax1.set_ylabel(cb.CB_DEFAULTS["Fluid_Defaults"]['Property_Names'][ordinate])
            ax1.plot(x, self.state_array_sf[:, ordinate], 'b--s', 
                     label='Secondary Fluid', markersize=5)

        
        # Add title, legend, and grid
        plt.title(cb.CB_DEFAULTS["Fluid_Defaults"]['Property_Names'][ordinate] \
                  + ' vs Area')
        fig.legend(loc="upper right", bbox_to_anchor=(1,1), 
                   bbox_transform=ax1.transAxes)
        ax1.grid(True)
        
        plt.show()
        
if __name__ == "__main__":
    
    #for simple calculation with A
    U_i = 250
    A = 0.12
    mdot_1, mdot_2 = 0.02, 0.04
    cells = 40 # resolution
    method = "Cavallini_Smith_Zecchin"
    
    #testcase 1. Attention! Refprop only works correctly, if restart console
    FLUID1 = "Propane * Pentane"
    comp1 = [.5, .5]
    fluid1= cb.fprop.init_fluid(FLUID1, comp1)
    st1_in = fluid1.set_state([380., 0.5], "TQ")
    
    FLUID2 = "water"
    comp2 = [1]
    fluid2 = cb.fprop.init_fluid(FLUID2, comp2)
    st2_in =  fluid2.set_state([300., 6e5], "TP", cb.fprop._TRANS_STRING)
    
    
    my_first_he =heat_exchanger([fluid1, fluid2], [st1_in, st2_in],
                                 [mdot_1, mdot_2], cells)
    my_first_he.geo_A(A)
    my_first_he.calc_NTU(U_i)
    my_first_he.diagram()
    
    #for calculation including heat transfer coefficients + geometry
    my_second_he = heat_exchanger([fluid1, fluid2], [st1_in, st2_in],
                                 [mdot_1, mdot_2], cells)

    my_second_he.geo_double_pipe(10e-3, 14e-3, 18e-3, 22e-3, 4, 15.4, method)
    st_2 = time.time()
    my_second_he.calc_NTU()
    et_2 = time.time()
    my_second_he.diagram()
    my_second_he.diagram(1, True)
    
    print(f"NTU-cell-method: {et_2-st_2} s")
    
    #for solve IVP
    
    my_third_he = heat_exchanger([fluid1, fluid2], [st1_in, st2_in],
                                 [mdot_1, mdot_2], cells)
    
    my_third_he.geo_double_pipe(10e-3, 14e-3, 18e-3, 22e-3, 4, 15.4, method)
    st_3 = time.time()
    my_third_he.calc_IVP()
    et_3 = time.time()
    my_third_he.diagram()
    my_third_he.diagram(1, True)
    
    print(f"solve_ivp-method: {et_3-st_3} s")
    
    # evaporator
    FLUID1 = "Propane * Pentane"
    comp1 = [.5, .5]
    fluid1= cb.fprop.init_fluid(FLUID1, comp1)
    st1_in = fluid1.set_state([296., 0.5], "TQ")
    
    FLUID2 = "water"
    comp2 = [1]
    fluid2 = cb.fprop.init_fluid(FLUID2, comp2)
    st2_in =  fluid2.set_state([360., 6e5], "TP", cb.fprop._TRANS_STRING)
    
    my_fourth_he = heat_exchanger([fluid1, fluid2], [st1_in, st2_in],
                                 [mdot_1, mdot_2], cells)
    my_fourth_he.geo_A(A)
    my_fourth_he.calc_NTU(U_i)
    my_fourth_he.diagram()
    my_fourth_he.diagram(1, True)
    
    # evaporator with NTU and alpha, dp
    method = "Chen_Bennett"
    
    my_fifth_he = heat_exchanger([fluid1, fluid2], [st1_in, st2_in],
                                 [mdot_1, mdot_2], cells)
    my_fifth_he.geo_double_pipe(10e-3, 14e-3, 18e-3, 22e-3, 4, 15.4, method)
    my_fifth_he.calc_NTU()
    my_fifth_he.diagram()
    my_fifth_he.diagram(1, True)
   
    # evaporator with solve ivp and alpha, dp
    
    my_sixth_he = heat_exchanger([fluid1, fluid2], [st1_in, st2_in],
                                 [mdot_1, mdot_2], cells)
    my_sixth_he.geo_double_pipe(10e-3, 14e-3, 18e-3, 22e-3, 4, 15.4, method)
    my_sixth_he.calc_IVP()
    my_sixth_he.diagram()
    my_sixth_he.diagram(1, True)
    
    # Graveyard
    # =============================================================================
    # SOLVE BVP NOT STABLE    
    #    def calc_BVP(self):
    #         x = np.linspace(0, self.l, self.resolution)
    #         y0 = [np.linspace(self.state_in_wf[2], self.state_in_wf[2], self.resolution),
    #               np.linspace(self.state_in_sf[2], self.state_in_sf[2], self.resolution),
    #               np.linspace(self.state_in_wf[1], self.state_in_wf[1], self.resolution),
    #               np.linspace(self.state_in_sf[1], self.state_in_sf[1], self.resolution)]
    # 
    #         res = solve_bvp(self._ODE_BVP, self._bc, x, y0, tol=1e-1, max_nodes=100)
    # 
    #         # Auslesen Berechnungsergebnisse
    #         h_wf_res = res.y[0]
    #         h_sf_res = res.y[1]
    #         p_wf_res = res.y[2]
    #         p_sf_res = res.y[3]
    # 
    #         # Berechnen des Temperaturverlaufes
    #         state_wf = []
    #         state_sf = []
    #         for l in range(len(h_wf_res)):
    #             state_wf.append(self.fluids[0].set_state([h_wf_res[l], p_wf_res[l]], "HP"))
    #             state_sf.append(self.fluids[1].set_state([h_sf_res, p_sf_res], "HP", cb.fprop._TRANS_STRING))
    #         self.state_array_wf = np.array(state_wf)
    #         self.state_array_sf = np.array(state_sf)
    #         
    #     def _bc(self, y_left, y_right):
    #         return np.array([y_left[0] - self.state_in_wf[2],
    #                          y_right[1] - self.state_in_sf[2],
    #                          y_left[2] - self.state_in_wf[1],
    #                          y_right[3] - self.state_in_sf[1]])
    #         
    #     def _ODE_BVP(self, x, y):
    #         h_wf = y[0]
    #         h_sf = y[1]
    #         p_wf = y[2]
    #         p_sf = y[3]
    #         state_wf_x = []
    #         state_sf_x = []
    #         R_ges = []
    #         dp_wf = []
    #         dp_sf = []
    #         
    #         for i in range(len(h_wf)):        
    #             state_wf_x.append(self.fluids[0].set_state([h_wf[i], p_wf[i]], "HP"))
    #             state_sf_x.append(self.fluids[1].set_state([h_sf[i], p_sf[i]], "HP", cb.fprop._TRANS_STRING))
    # 
    #             R_gesi, dp_wfi, dp_sfi= self.thermal_resistance(state_wf_x[i], state_sf_x[i])
    #             R_ges.append(R_gesi)
    #             dp_wf.append(-dp_wfi)
    #             dp_sf.append(dp_sfi)
    #         state_wf_x = np.array(state_wf_x)
    #         state_sf_x = np.array(state_sf_x)
    #         R_ges = np.array(R_ges)
    #         dp_wf = np.array(dp_wf)
    #         dp_sf= np.array(dp_sf)
    #         delta_T = state_sf_x[:,0] - state_wf_x[:,0]   
    # 
    #         dh_wf = self.di * np.pi/R_ges * delta_T / self.mdot_wf
    #         dh_sf = self.di * np.pi/R_ges * delta_T / self.mdot_sf
    # 
    #         dy = np.array([dh_wf, dh_sf, dp_wf, dp_sf])
    #         #print(y)
    #         return dy
    # =============================================================================
    
    



# -*- coding: utf-8 -*-
"""
Created on Mon Aug 12 17:39:12 2024

@author: atakan
"""
import numpy as np
import carbatpy as cb
import scipy.optimize as opti


dir_name = cb.CB_DEFAULTS["General"]["CB_DATA"]+"\\io-cycle-data.yaml"

def _change_h_w_out(hex_act, h_out_new):
    inputs = hex_act.inputs["states"]
    out_state = {}
    in_state = {}
    for flname, value in inputs.items():
        out_state[flname] = inputs[flname]['out']
        in_state[flname] = inputs[flname]['in']
    out_state["working_fluid"][2] = h_out_new
    return in_state, out_state

def _opti_h_func( h_act_out, hex_act,  m_dot_w, verbose =False):

    run_p_cond = {"m_dot": {"working_fluid": m_dot_w}}
    instates, outstates =_change_h_w_out(hex_act, h_act_out)
    hex_act.calculate(instates, outstates,run_param=run_p_cond)
    if verbose:
        print (hex_act.output["dt_mean"],  hex_act.warning)
    if hex_act.warning.value > 2:
        return +100
    return np.abs(hex_act.output["dt_mean"])

def hex_opti_work_out(self, run_p_cond, verbose =False):
    self.calculate(run_param=run_p_cond)
    if verbose:
        print(f"opti_work_out: T-mean before optimization= {condenser.output['dt_mean']} K")
    # condenser._change_h_w_out =_change_h_w_out
    tolerance = 1e-2
    max_iter = 240
    h_act = self.output["state_out"]["working_fluid"][2] *1.005
    dh = h_act *.051
    bound_act = opti.Bounds(lb =h_act-dh, ub = h_act+dh)

    result = opti.minimize(_opti_h_func,
                       h_act,
                       args=(
                             self,
                             m_dot_w),
                       method='Nelder-Mead',
                       tol=tolerance,
                       bounds=bound_act,
                       options={"maxiter": max_iter, # can take long!
                                "disp": True})
    if verbose:
        print(f"T-mean before optimization: {result.fun} K\n{result}")
    

# ================ CALCULATIONS ==============================
# ------ Start/initial condition ----
# but the mass flow rate is yet unknown, plotting must be delayed
start = cb.comp.Start("start", dir_name, m_dot=10e-3)

# ----- compressor --------------
# prescribed power, working_fluid mass flow rate is calculated here
run_p_comp = {"power": 1500.}
compressor = cb.comp.FlowMachine("compressor", dir_name)
p_high = compressor.config['working_fluid']['p_high']
compressor.calculate(start.output["state_out"],
                     {'working_fluid': [600, p_high, 5e5]},
                     run_param=run_p_comp)  # ,  m_dot=10e-3)
# for the output only p_high is used! Now m_dot is known for the working fluid.
m_dot_w = compressor.output["m_dot"]
start = cb.comp.Start("start", dir_name, m_dot=m_dot_w)

# ----- coondenser --------------
run_p_cond = {"m_dot": {"working_fluid": m_dot_w}}

condenser = cb.comp.StaticHeatExchanger("condenser", dir_name)

condenser.calculate(run_param=run_p_cond)
print(f"T-mean before optimization: {condenser.output['dt_mean']} K")
# condenser._change_h_w_out =_change_h_w_out
tolerance = 1e-2
max_iter =140
h_act = condenser.output["state_out"]["working_fluid"][2] *1.005
dh = h_act *.021
bound_act = opti.Bounds(lb =h_act-dh, ub = h_act+dh)

result = opti.minimize(_opti_h_func,
                   h_act,
                   args=(
                         condenser,
                         m_dot_w),
                   method='Nelder-Mead',
                   tol=tolerance,
                   bounds=bound_act,
                   options={"maxiter": max_iter, # can take long!
                            "disp": True})
print(f"T-mean before optimization: {result.fun} K\n{result}")


# -*- coding: utf-8 -*-
"""
Example how to write flexible to_dir, from_dir, save_to_file, load_from_file for classes.

Also for other instances and numpy arrays.

From ChatGPT

Created on Thu Jul 25 14:13:52 2024

@author: atakan
"""

import json
import yaml
import numpy as np

class AnotherClass:
    def __init__(self, value):
        self.value = value

    def to_dict(self):
        return {'value': self.value}

    @classmethod
    def from_dict(cls, data):
        return cls(data['value'])

class MyClass:
    def __init__(self, name, age, occupation, array, component):
        self.name = name
        self.age = age
        self.occupation = occupation
        self.array = array
        self.component = component  # Beliebig benannter Attribut

    def to_dict(self):
        # Alle Attribute extrahieren, keine Methoden und self.all ausschließen
        result = {}
        for key, value in self.__dict__.items():
            if key != 'all':  # Beispiel für ein zu ignorierendes Attribut
                if isinstance(value, np.ndarray):
                    result[key] = value.tolist()  # NumPy-Arrays in Listen konvertieren
                elif hasattr(value, 'to_dict'):
                    result[key] = value.to_dict()  # Konvertieren, wenn `to_dict` vorhanden
                else:
                    result[key] = value
        return result

    @classmethod
    def from_dict(cls, data):
        instance = cls.__new__(cls)
        for key, value in data.items():
            if isinstance(value, list) and len(value) > 0 and isinstance(value[0], (int, float)):
                # Annahme: Listen sind NumPy-Arrays
                setattr(instance, key, np.array(value))
            elif isinstance(value, dict) and 'value' in value:
                # Annahme: Dictionary ist eine Instanz von AnotherClass
                setattr(instance, key, AnotherClass.from_dict(value))
            else:
                setattr(instance, key, value)
        return instance

    def save_to_file(self, file_path):
        try:
            data = self.to_dict()
            if file_path.endswith('.json'):
                with open(file_path, 'w') as file:
                    json.dump(data, file, indent=4)
            elif file_path.endswith('.yaml') or file_path.endswith('.yml'):
                with open(file_path, 'w') as file:
                    yaml.dump(data, file)
            else:
                raise ValueError("Unsupported file format. Use .json or .yaml/.yml")
        except Exception as e:
            print(f"Error saving to file: {e}")

    @classmethod
    def load_from_file(cls, file_path):
        try:
            if file_path.endswith('.json'):
                with open(file_path, 'r') as file:
                    data = json.load(file)
            elif file_path.endswith('.yaml') or file_path.endswith('.yml'):
                with open(file_path, 'r') as file:
                    data = yaml.load(file, Loader=yaml.SafeLoader)
            else:
                raise ValueError("Unsupported file format. Use .json or .yaml/.yml")

            return cls.from_dict(data)
        except Exception as e:
            print(f"Error loading from file: {e}")
            return None
     
        
if __name__ =="__main__":
    # Beispielinstanzen erstellen
    array = np.array([1, 2, 3, 4])
    component = AnotherClass(10)
    obj = MyClass("Alice", 30, "Engineer", array, component)
    
    # Speichern der Instanz in eine JSON-Datei
    obj.save_to_file('instance.json')
    
    # Neue Instanz aus der Datei erstellen
    new_obj = MyClass.load_from_file('instance.json')
    if new_obj:
        print(new_obj.__dict__)



# -*- coding: utf-8 -*-
"""
Created on Fri Aug  2 15:28:50 2024

@author: atakan
"""
class StaticHeatExchanger:
    def __init__(self, fluids, h_dot_min, h_out_w, h_limit_s, **kwargs):
        self.fluids = fluids
        state_in_w = fluids[0].properties.state
        self.m_dot_s = 0
        self.h_limit_s = h_limit_s
        self.q_dot = h_dot_min
        self.m_dot_w = np.abs(h_dot_min / (h_out_w - state_in_w[2]))
        self.h_out_w = h_out_w
        self.h_out_s = float(copy.copy(h_limit_s))
        self.points = kwargs.get('points', 50)
        self.d_temp_separation_min = kwargs.get('d_temp_separation_min', 0.5)
        self.calc_type = kwargs.get('calc_type', "const")
        self.name = kwargs.get('name', "evaporator")
        self.plot_info = kwargs.get('plot_info', None)
        self.heating = -1
        if h_out_w < state_in_w[2]:
            self.heating = 1
        self.all_states = np.zeros(
            (2, self.points, len(cb.fprop._THERMO_STRING.split(";"))))
        self.h_in_out = np.zeros((2, 4))
        self.dt_mean = None
        self.dt_min = None
        self.dt_max = None
        self.warning = 0
        self.warning_message = "All o.k."

    @classmethod
    def create_from_input(cls, input_hex, temp_ambient, exchanger_name):
        # Find the working fluid and storage fluid details
        working_fluid_info = input_hex.get("working_fluid")
        storage_fluid_info = None

        for key, value in input_hex.items():
            if "storage" in key.lower():
                storage_fluid_info = value
                break

        if not working_fluid_info or not storage_fluid_info:
            raise ValueError("Missing working fluid or storage fluid information in input data.")

        # Initialize fluids
        working_fluid = init_fluid(working_fluid_info)
        storage_fluid = init_fluid(storage_fluid_info)

        # Determine the species mappings for the given exchanger
        species_mapping = input_hex[exchanger_name]["species"]
        working_fluid_states = species_mapping["working_fluid"]
        storage_fluid_states = next(value for key, value in species_mapping.items() if "storage" in key.lower())

        # Extract temperatures and pressures for storage fluid
        
        temp_out_storage = storage_fluid_info[storage_fluid_states["out"]]
        temp_in_storage = storage_fluid_info[storage_fluid_states["in"]]
        if temp_in_storage == "ambient":
            temp_in_storage = temp_ambient
        if temp_out_storage == "ambient":
            temp_out_storage = temp_ambient

        p_low_storage = storage_fluid_info["p_low"]

        # Calculate enthalpies for storage fluid
        
        state_out_storage = storage_fluid.set_state([temp_out_storage, p_low_storage], "TP")
        state_in_storage = storage_fluid.set_state([temp_in_storage, p_low_storage], "TP")

        h_in_storage = state_in_storage[2]
        h_out_storage = state_out_storage[2]

        # Determine direction of heat exchange
        dt_min = input_hex[exchanger_name]["dt_min"]
        temp_in_working = working_fluid_info[working_fluid_states["in"]]
        temp_out_working = working_fluid_info[working_fluid_states["out"]]
        if temp_in_working == "ambient":
            temp_in_working = temp_ambient
        if temp_out_working == "ambient":
            temp_out_working = temp_ambient

        if temp_in_working > temp_in_storage:
            heating = 1  # Evaporator (heating of secondary fluid)
            temp_in_working += dt_min
            temp_out_working += dt_min
        else:
            heating = -1  # Condenser
            temp_in_working -= dt_min
            temp_out_working -= dt_min

        # Calculate enthalpies for working fluid
        p_low_working = working_fluid_info["p_low"]
        state_in_working = working_fluid.set_state([temp_in_working, p_low_working], "TP")
        state_out_working = working_fluid.set_state([temp_out_working, p_low_working], "TP")

        h_in_working = state_in_working[2]
        h_out_working = state_out_working[2]

        # Create StaticHeatExchanger instance
        return cls(
            fluids=[working_fluid, storage_fluid],
            h_dot_min=input_hex[exchanger_name]["q_dot"],
            h_out_w=h_out_working,
            h_limit_s=h_out_storage,
            d_temp_separation_min=dt_min,
            calc_type=input_hex[exchanger_name].get("calc_type", "const"),
            points=input_hex[exchanger_name].get("points", 50),
            name=input_hex[exchanger_name].get("name", exchanger_name),
            plot_info=input_hex[exchanger_name].get("plot_info", None),
        )

# Example usage
temp_ambient = 293.15  # Example ambient temperature in Kelvin
input_hex = {
    'condenser': {
        'model': "StaticHeatExchanger",
        "calc_type": "const",
        "species": {"working_fluid": {"in": "temp_high", "out": "temp_low"},
                    "hot_storage": {"in": "temp_low", "out": "temp_high"}},
        "dt_min": 3.,
        "q_dot": 1000.,
        "overall_u": 100,
        'area': 1,
    },
    'hot_storage': {
        "species": "Water",
        "fractions": [1.0],
        "p_low": 2e5,
        "temp_low": "ambient",
        "temp_high": 360.0,
        'props': "REFPROP",
    },
    'working_fluid': {
        "species": "Propane * Butane * Pentane * Hexane",
        "fractions": [
            7.71733787e-01,
            2.22759488e-02,
            1.78685867e-01,
            0.027304397199999997],
        "temp_low": "ambient",
        "temp_high": 360.0,
        "p_low": 1.28250708e+05,
        "p_high": 1.37548728e+06,
        "optimize": "None",
        "setting": "initial",
        'props': "REFPROP",
    },
}

heat_exchanger_instance = StaticHeatExchanger.create_from_input(input_hex, temp_ambient, "condenser")



# -*- coding: utf-8 -*-
"""
Created on Wed Nov 27 12:54:16 2024

@author: atakan
"""

"""Thank you for the clarification! Let's implement a Python function that handles the three cases you described. We'll use conditional statements to determine which values are given and then perform the appropriate calculations.

Here's a Python function that handles the three cases:

```python"""
def calculate_thermodynamic_values(Q=None, m_w=None, h_w_e=None, h_w_a=None, m_s=None, h_s_e=None, h_s_a=None):
    if Q is not None and m_w is not None and h_s_a is not None:
        # Case 1: Q, m_w, and h_s_a are given
        h_w_a = h_w_e - Q / m_w
        m_s = Q / (h_s_a - h_s_e)
        return {
            'h_w_a': h_w_a,
            'm_s': m_s
        }
    elif Q is not None and h_w_a is not None and h_s_a is not None:
        # Case 2: Q, h_w_a, and h_s_a are given
        m_w = Q / (h_w_e - h_w_a)
        m_s = Q / (h_s_a - h_s_e)
        return {
            'm_w': m_w,
            'm_s': m_s
        }
    elif m_w is not None and h_w_a is not None and h_s_a is not None:
        # Case 3: m_w, h_w_a, and h_s_a are given
        Q = m_w * (h_w_e - h_w_a)
        m_s = Q / (h_s_a - h_s_e)
        return {
            'Q': Q,
            'm_s': m_s
        }
    else:
        raise ValueError("Insufficient or incorrect input values provided.")

# Example usage:
# Case 1
result_case1 = calculate_thermodynamic_values(Q=1000, m_w=5, h_w_e=200, h_s_e=150, h_s_a=300)
print("Case 1:", result_case1)

# Case 2
result_case2 = calculate_thermodynamic_values(Q=1000, h_w_e=200, h_w_a=100, h_s_e=150, h_s_a=300)
print("Case 2:", result_case2)

# Case 3
result_case3 = calculate_thermodynamic_values(m_w=5, h_w_e=200, h_w_a=100, h_s_e=150, h_s_a=300)
print("Case 3:", result_case3)
""" 

### Explanation:
1. **Case 1**: If \( Q \), \( m_w \), and \( h_{s,a} \) are given, we calculate \( h_{w,a} \) and \( m_s \).
2. **Case 2**: If \( Q \), \( h_{w,a} \), and \( h_{s,a} \) are given, we calculate \( m_w \) and \( m_s \).
3. **Case 3**: If \( m_w \), \( h_{w,a} \), and \( h_{s,a} \) are given, we calculate \( Q \) and \( m_s \).

The function returns a dictionary with the calculated values for each case. If the input values do not match any of the cases, it raises a `ValueError`.

You can adjust the example usage to fit your specific input values and test the function accordingly.
"""


# -*- coding: utf-8 -*-
"""
Created on Fri Aug  2 15:01:26 2024

@author: atakan
"""
import carbatpy as cb


def create_heat_exchanger_from_input(input_hex, temp_ambient, exchanger_name):
    # Find the working fluid and storage fluid details
    # Find the working fluid and storage fluid details
    working_fluid_info = input_hex.get("working_fluid")
    storage_fluid_info = None

    for key, value in input_hex.items():
        if "storage" in key.lower():
            storage_fluid_info = value
            break

    if not working_fluid_info or not storage_fluid_info:
        raise ValueError(
            "Missing working fluid or storage fluid information in input data.")

    # Initialize fluids
    working_fluid = cb.init_fluid(working_fluid_info["species"],
                                  working_fluid_info["fractions"],
                                  props=working_fluid_info["props"]
                                  )
    storage_fluid = cb.init_fluid(storage_fluid_info["species"],
                                  storage_fluid_info["fractions"],
                                  props=storage_fluid_info["props"])

    # Determine the species mappings for the given exchanger
    species_mapping = input_hex[exchanger_name]["species"]
    working_fluid_states = species_mapping["working_fluid"]
    storage_fluid_states = next(
        value for key, value in species_mapping.items() if "storage" in key.lower())

    # Extract temperatures and pressures for storage fluid

    temp_out_storage = storage_fluid_info[storage_fluid_states["out"]]
    temp_in_storage = storage_fluid_info[storage_fluid_states["in"]]
    if temp_in_storage == "ambient":
        temp_in_storage = temp_ambient
    if temp_out_storage == "ambient":
        temp_out_storage = temp_ambient

    p_low_storage = storage_fluid_info["p_low"]

    # Calculate enthalpies for storage fluid

    state_out_storage = storage_fluid.set_state([temp_out_storage, p_low_storage], "TP")
    state_in_storage = storage_fluid.set_state([temp_in_storage, p_low_storage], "TP")

    h_in_storage = state_in_storage[2]
    h_out_storage = state_out_storage[2]

    # Determine direction of heat exchange
    dt_min = input_hex[exchanger_name]["dt_min"]
    temp_in_working = working_fluid_info[working_fluid_states["in"]]
    temp_out_working = working_fluid_info[working_fluid_states["out"]]
    if temp_in_working == "ambient":
        temp_in_working = temp_ambient
    if temp_out_working == "ambient":
        temp_out_working = temp_ambient

    if temp_in_working > temp_in_storage:
        heating = 1  # Evaporator (heating of secondary fluid)
        temp_in_working += dt_min
        temp_out_working += dt_min
    else:
        heating = -1  # Condenser
        temp_in_working -= dt_min
        temp_out_working -= dt_min

    # Calculate enthalpies for working fluid
    p_low_working = working_fluid_info["p_low"]
    
    state_out_working = working_fluid.set_state([temp_out_working, p_low_working], "TP")
    state_in_working = working_fluid.set_state([temp_in_working, p_low_working], "TP")
    h_in_working = state_in_working[2]
    h_out_working = state_out_working[2]

    # Create StaticHeatExchanger instance
    heat_exchanger = cb.hex_th.StaticHeatExchanger(
        fluids=[working_fluid, storage_fluid],
        h_dot_min=input_hex[exchanger_name]["q_dot"],
        h_out_w=h_out_working,
        h_limit_s=h_out_storage,
        d_temp_separation_min=dt_min,
        calc_type=input_hex[exchanger_name].get("calc_type", "const"),
        points=input_hex[exchanger_name].get("points", 50),
        name=input_hex[exchanger_name].get("name", exchanger_name),
        plot_info=input_hex[exchanger_name].get("plot_info", None),
    )

    return heat_exchanger


# Example usage
temp_ambient = 283.15  # Example ambient temperature in Kelvin
input_hex = {
    'condenser': {
        'model': "StaticHeatExchanger",
        "calc_type": "const",
        "species": {"working_fluid": {"in": "temp_high", "out": "temp_low"},
                    "hot_storage": {"in": "temp_low", "out": "temp_high"}},
        "dt_min": 3.,
        "q_dot": 1000.,
        "overall_u": 100,
        'area': 1,
    },
    'hot_storage': {
        "species": "Water",
        "fractions": [1.0],
        "p_low": 2e5,
        "temp_low": "ambient",
        "temp_high": 370.0,
        'props': "REFPROP",
    },
    'working_fluid': {
        "species": "Propane * Butane * Pentane * Hexane",
        "fractions": [
            7.71733787e-01,
            2.22759488e-02,
            1.78685867e-01,
            0.027304397199999997],
        "temp_low": "ambient",
        "temp_high": 420.0,
        "p_low": 1.128250708e+05,
        "p_high": 1.37548728e+06,
        "optimize": "None",
        "setting": "initial",
        'props': "REFPROP",
    },
}

heat_exchanger_instance = create_heat_exchanger_from_input(
    input_hex, temp_ambient, "condenser")
heat_exchanger_instance.pinch_calc()
heat_exchanger_instance.pinch_plot()
if heat_exchanger_instance.warning:
    print(heat_exchanger_instance.warning_message)



# -*- coding: utf-8 -*-
"""
Created on Sat Aug  3 17:36:05 2024

@author: atakan
"""

import carbatpy as cb
from carbatpy.helpers.ser_dict import Serializable, DataAll, DataNode
import json
import yaml
import numpy as np
import matplotlib.pyplot as plt
import copy
from scipy.optimize import minimize


class HeatExchangerBase(Serializable):
    """Base class for heat exchangers.

    All inputs come either from a file (json/yaml) or from a dictionary.

    """

    def __init__(self, inputs, name=None, **kwargs):
        """
        Set the heat exchanger input parameters.

        Parameters
        ----------
        inputs : string or dict
            If string, it must be a valid filename (json or yaml) with the dictionary content.
        name : string, optional
            The name of the heat exchanger. The default is None.
        **kwargs : dict
            - plot_info : dict, a dictionary with file, axes (from matplotlib), labels, symbols
              used for plotting (as defined in self.plot()).

        Raises
        ------
        NotImplementedError
            If the inputs are not correctly provided.

        """
        self.name = name if name else "heat-exchanger"
        self.plot_info = kwargs.get("plot_info", None)
        self.output = Serializable()
        self.all_states = {}
        self.warning = DataAll(value=0, message="All o.k.")
        self._defaults = DataAll()

        # Load inputs from a file or directly from a dictionary
        print("hier", isinstance(inputs, dict), inputs, name)
        if isinstance(inputs, dict):
            self.inputs = inputs
            print("hier")
        elif isinstance(inputs, str):
            try:
                with open(inputs, "r") as file:
                    if "json" in inputs:
                        self.inputs = json.load(file)
                    elif "yaml" in inputs:
                        self.inputs = yaml.safe_load(file)
                    else:
                        raise ValueError("Unsupported file format.")
            except Exception as e:
                print(f"Error loading input file: {e}")
                raise
        else:
            raise NotImplementedError(
                f"Unsupported input type: {type(inputs)}, {inputs}"
            )

        self._inp = DataAll(**self.inputs)

        match self.inputs[self.name]["model"]:  # noqa
            case "StaticHeatExchanger":
                pass
            case _:
                raise NotImplementedError(
                    f"This heat exchanger type is not implemented: {self._inp.process[self.name]}")

    def calc(self, **kwargs):
        """Selects the calculation method and calls the appropriate function."""
        raise NotImplementedError("This method must be overridden in subclasses.")

    def plot(self, **kwargs):
        """Plot the results into a given or a new figure."""
        raise NotImplementedError("This method must be overridden in subclasses.")

    def save(self, filename, **kwargs):
        """Save all results into one or several files."""
        raise NotImplementedError("This method must be overridden in subclasses.")

    def _costs(self, filename, **kwargs):
        """Calculate costs via calc()."""
        raise NotImplementedError("This method must be overridden in subclasses.")


class StaticHeatExchanger(HeatExchangerBase):
    """Class for static counter-flow heat exchanger.

    No time dependence and no heat transfer coefficients * areas are used (UA)!
    Instead, a minimum approach temperature is tried to be met.
    """

    # fluids, h_dot_min, h_out_w, h_limit_s, **kwargs):
    def __init__(self, inputs, name=None, **kwargs):
        """
        Initialize a static heat exchanger for steady-state calculations.

        Parameters
        ----------
        fluids : list of cb.fprop.Fluid
            The definition of the two fluids, as they enter the heat exchangers.
        h_dot_min : float
            Enthalpy flow rate (W) which has to be transferred.
        h_out_w : float
            Exit enthalpy of the working fluid.
        h_limit_s : float
            Limit in enthalpy for the secondary fluid.
        kwargs : dict, optional
            Optional parameters for the heat exchanger configuration.
        """
        super().__init__(inputs, name, **kwargs)
        fluids, h_dot_min, h_out_w, h_limit_s, d_temp_separation = from_dict(inputs, name, kwargs)
        self.fluids = fluids
        state_in_w = fluids[0].properties.state
        self.m_dot_s = 0
        self.h_limit_s = h_limit_s
        self.q_dot = h_dot_min
        self.m_dot_w = np.abs(h_dot_min / (h_out_w - state_in_w[2]))
        self.h_out_w = h_out_w
        self.h_out_s = float(copy.copy(h_limit_s))
        self.points = kwargs.get('points', 50)
        self.d_temp_separation_min = kwargs.get('d_temp_separation_min', 0.5)
        self.calc_type = kwargs.get('calc_type', "const")
        self.name = kwargs.get('name', "evaporator")
        self.plot_info = kwargs.get('plot_info', None)
        self.heating = 1 if h_out_w < state_in_w[2] else -1
        self.all_states = np.zeros(
            (2, self.points, len(cb.fprop._THERMO_STRING.split(";"))))
        self.h_in_out = np.zeros((2, 4))
        self.dt_mean = None
        self.dt_min = None
        self.dt_max = None
        self.warning = 0
        self.warning_message = "All o.k."

    @property
    def plot_info(self):
        return self._plot_info

    @plot_info.setter
    def plot_info(self, value):
        if value is None:
            value = {}
        elif not isinstance(value, dict):
            raise ValueError("plot_info must be a dictionary")
        self._plot_info = value

    @property
    def all(self):
        return {key: value for key, value in self.__dict__.items() if key != 'all'}

    def to_dict(self):
        result = {}
        for key, value in self.__dict__.items():
            if key != 'all':
                if isinstance(value, np.ndarray):
                    result[key] = value.tolist()
                elif hasattr(value, 'to_dict'):
                    result[key] = value.to_dict()
                else:
                    result[key] = value
        return result

    
    def pinch_calc(self,  verbose=False):
        """
        Calculate the changes in enthalpy and temperature in the heat exchanger

        counter-flow hex assumed! Both flows are isobaric.
        Is used to check, whether the second law is violated. The factor can
        be used to vary the mass flow rate of the working fluid, until no
        violation is found (done in root finding).

        Parameters
        ----------
        verbose : boolean, optional
            if True, several variables will be printed out. Default is False.



        Raises
        ------
        Exception
            if temperatures are not consistent.

        Returns
        -------
        m_dot_s : float
            secondary fluid mass flow rate (kg/s.
        d_tempall : numpy-array
            The temperature differences along the counter-flow heat exchanger.
        w_array : array
            properties of the working fluid along the heat exchanger
            (T,p,h, etc. see fluid class).
        s_array : array
            properties of the secondary fluid along the heat exchanger
            (T,p,h, etc. see fluid class).

        """
        self.warning = 0
        self.warning_message = "All o.k."
        w_in = copy.copy(self.fluids[0])
        s_in = copy.copy(self.fluids[1])

        w_out = copy.copy(self.fluids[0])  # not yet the correct state!
        s_out = copy.copy(self.fluids[1])

        #  fixed values
        self. h_in_out[1, 0] = s_in.properties.enthalpy
        self. h_in_out[0, 0] = w_in.properties.enthalpy
        self. h_in_out[0, 1] = self.h_out_w

        # fixed limiting state, secondary fluid
        state_out_s = s_out.set_state([self.h_out_s,
                                       s_in.properties.pressure], "HP")

        self. h_in_out[1, 1] = state_out_s[2]

        h_delta_s = np.abs(
            state_out_s[2] - s_in.properties.enthalpy)

        # fixed heat flow,  determines mass flow rate
        self.m_dot_s = self.q_dot / h_delta_s

        s_array = self._calculate_state_array(s_out, self.h_in_out[1, :2])
        w_array = self._calculate_state_array(w_out, self.h_in_out[0, 1::-1])
        # temperature difference, ok?

        d_tempall = w_array[:, 0]-s_array[:, 0]

        self.dt_mean, self.dt_min, self.dt_max = d_tempall.mean(), np.abs(
            d_tempall).min(), np.abs(d_tempall).max()

        self._check_temperature_consistency(d_tempall)

        if self.plot_info:
            self._plot_heat_exchanger(w_array, s_array)

        if verbose:
            self._print_verbose(d_tempall)
        self.all_states[0,:,:] = w_array
        self.all_states[1,:,:] = s_array
        return self.m_dot_s, d_tempall*self.heating, w_array, s_array

    def _pinch_root(self, h_out_s,  secondary, verbose=False):
        """
        Function for finding the minimum mean temperature difference in a heat
        exchanger, while not going below the minimum approach temperature.

        The output enthalpy of the secondary fluid is set (as default) and the
        heat exchanger is evaluated.


        Parameters
        ----------
        h_out_s : float
            the output enthalpy, default would be of the secondary fluid.
        secondary : boolean
            shall the secondary fluid output be varied (or the working fluid)?

        Returns
        -------
        mean T-difference, float
            root tries to reach a value of 0.

        """

        if isinstance(h_out_s, float):  # numpy expects a float (later)
            value = h_out_s
        else:
            value = h_out_s[0]

        if secondary:
            self.h_out_s = value
        else:
            self.h_out_w = value

        mdot_s, d_temps, wf_states, sf_states = self.pinch_calc()

        if self.warning < 100:
            return abs(self.dt_mean)
        return 500.0
    
    def _p_opti_help(self, values, secondary=True, verbose=False, **kwargs):
        what =kwargs.get("what","TP")
        pressure_w, h_out = values
        name_o = what.replace("P","")
        if verbose:
            print("p_opti:", values, name_o,self.dt_mean)
        if name_o == "T":
            other =self.fluids[0].val_dict["Temperature"]
        else:
            raise NotImplementedError(f"pressure optimization not implemnted for {name_o}")
        
        state_in = self.fluids[0].set_state([pressure_w, other], what)
        self.find_pinch(secondary)
        if verbose:
            print("p_opti:", values, name_o,self.dt_mean)
        if self.warning ==0:
            return self.dt_mean
        return 999.0

    def _pinch_root(self, h_out_s,  secondary, verbose=False):
        """
        Function for finding the minimum mean temperature difference in a heat
        exchanger, while not going below the minimum approach temperature.

        The output enthalpy of the secondary fluid is set (as default) and the
        heat exchanger is evaluated.


        Parameters
        ----------
        h_out_s : float
            the output enthalpy, default would be of the secondary fluid.
        secondary : boolean
            shall the secondary fluid output be varied (or the working fluid)?

        Returns
        -------
        mean T-difference, float
            root tries to reach a value of 0.

        """

        if isinstance(h_out_s, float):  # numpy expects a float (later)
            value = h_out_s
        else:
            value = h_out_s[0]

        if secondary:
            self.h_out_s = value
        else:
            self.h_out_w = value

        mdot_s, d_temps, wf_states, sf_states = self.pinch_calc()

        if self.warning < 100:
            return abs(self.dt_mean)
        return 500.0
    def _handle_exception(self, e):
        print("find pinch:", type(e))
        print(e.args)
        print(e)
        print("root-exception", self.heating)
        
    def find_pinch(self, secondary=True, verbose = False):
        """
        Function tries to vary the secondary fluid enthalpy  until a
        minimum approach temperature is reached. This also changes the
        mass flow rate.  This is then also the new
        exit state
        within the heat exchanger. If this is also not succesful,
        self.warning is set to 1. This should be checked.
        Parameters
        ----------
        secondary : Boolean, optional
            shall the output state of the secondary (True) or the working
            (False) fluid be varied? default = True

        Returns
        -------
        float
            the optimized enthalpy of the secondary fluid.

        """
        
        x0 = copy.copy(self.h_out_s if secondary else self.h_out_w)
        # if secondary:
        #     x0 = copy.copy(self.h_out_s)
        # else:
        #     x0 = copy.copy(self.h_out_w)

        tolerance = 1e-3

        try:

            result = minimize(self._pinch_root, x0, args=(secondary,),
                              method='Nelder-Mead', tol=tolerance)

            if verbose:
                print(
                    f"result {result}, heating {self.heating}")

            if result.success or result.status == 2:
                if result.status == 2:
                    self.warning = 2  # T-difference probably smaller
                    self.warning_message = "Minimization problem: "+result.message
                if secondary:
                    self.h_out_s = result.x[0]
                else:
                    self.h_out_w = result.x[0]
                if verbose:
                    print(f"Min T-distance {self.dt_min:.3f}, Mean T-distance {self.dt_mean:.3f}")

                return result.x[0]

        # except:
        except Exception as e:
            self._handle_exception(e)
            return 10000
        print("root-finding problem! (in heat_exchanger_thermo_v2.find_pinch)",
              result)
        print(f"Heating: {self.heating}")
        self.warning = 1
        return self.warning
    
    def pinch_plot(self, plot_fname="", plotting=True, **kwargs):
        """
        calculates the secondary fluid output state and mass flow, for the
        minimum approach temperature of the HeatExchanger instance. When wanted,
        this is also plotted

        Parameters
        ----------
        plot_fname : string, optional
            file-name to store the plot. The default is "".
        plotting : Boolean, optional
            should it be plotted? The default is True.

        Returns
        -------
        m_dot_s : float
            mass flow raete of the secondary fluid in SI units (kg/s).
        d_tempall : np.array
            the temperature differences between the two fluids along the heat
            exchanger.
        w_array : np.array [self.points, 7]
            the states of the working fluid along the heat exchanger.
        s_array : np.array [self.points, 7]
            the states of the secondary fluid along the heat exchanger.

        """
        verbose =kwargs.get("verbose",False)
        if verbose:
            print(f"------pinch-plot running -----plot:{plotting}")
        m_dot_s, d_tempall, w_array, s_array = self.pinch_calc()

        if plotting:
            h_w_plot_array = (
                w_array[:, 2] - w_array[:, 2].min()) * self.m_dot_w
            fig, ax_one = plt.subplots(1, 1)
            ax_one.plot((s_array[:, 2] - s_array[:, 2].min()) * self.m_dot_s,
                        s_array[:, 0], "v")
            ax_one.plot(h_w_plot_array, w_array[:, 0], "o")
            ax_one.set_xlabel(
                "specific enthalpy flow per mass of secondary fluid / (J / kg)")
            ax_one.set_ylabel("temperature / (K)")
            ax_one.set_title("heat exchanger, simple")
        if plot_fname != "":
            fig.savefig(plot_fname)

        return m_dot_s, d_tempall, w_array, s_array
    
    def _p_opti_help(self, values, secondary=True, verbose=False, **kwargs):
        what =kwargs.get("what","TP")
        pressure_w, h_out = values
        name_o = what.replace("P","")
        if verbose:
            print("p_opti:", values, name_o,self.dt_mean)
        if name_o == "T":
            other =self.fluids[0].val_dict["Temperature"]
        else:
            raise NotImplementedError(f"pressure optimization not implemnted for {name_o}")
        
        state_in = self.fluids[0].set_state([pressure_w, other], what)
        self.find_pinch(secondary)
        if verbose:
            print("p_opti:", values, name_o,self.dt_mean)
        if self.warning ==0:
            return self.dt_mean
        return 999.0

    # def calculate_pinch_point(self, verbose=False):
    #     """Calculate pinch point and all other states for both fluids."""

    #     from scipy.optimize import fsolve

    #     results = self.all_states.copy()
    #     # initial guess
    #     if verbose:
    #         print("Guessing the enthalpy outlet for the secondary fluid")
    #     x0 = self.h_out_s
    #     if not x0:
    #         x0 = self.h_limit_s

    #     def _res_func(h_out_s):
    #         return self.find_pinch_root(h_out_s, verbose=verbose)[0, -1, 2] - self.h_limit_s

    #     self.h_out_s = fsolve(_res_func, x0)[0]
    #     if verbose:
    #         print(f"solved h_out_s = {self.h_out_s}")
    #     # set all states:
    #     self.all_states = self.find_pinch_root(self.h_out_s, verbose=verbose)

    # def find_optimal_pinch(self, secondary=True, verbose=False):
    #     """Find the optimal pinch point for the heat exchanger.

    #     Parameters
    #     ----------
    #     secondary : bool, optional
    #         If True, the optimization is for the secondary fluid. The default is True.
    #     verbose : bool, optional
    #         If True, additional information is printed. The default is False.

    #     Returns
    #     -------
    #     None.
    #     """
    #     from scipy.optimize import minimize

    #     # Define the objective function
    #     def objective(enthalpy_out):
    #         return np.abs(
    #             self.find_pinch_root(enthalpy_out, secondary, verbose)[0, -1, 2]
    #             - self.h_limit_s
    #         )

    #     # Initial guess for the outlet enthalpy
    #     initial_guess = self.h_out_s if secondary else self.h_out_w

    #     # Perform the optimization
    #     result = minimize(objective, initial_guess, method='Nelder-Mead')
    #     optimal_enthalpy_out = result.x[0]

    #     if secondary:
    #         self.h_out_s = optimal_enthalpy_out
    #     else:
    #         self.h_out_w = optimal_enthalpy_out

    #     self.calculate_pinch_point(verbose)

    # def optimize_pressure(self, values, secondary=True, verbose=False, **kwargs):
    #     """
    #     Optimizes the pressures and the mass flows of both fluids.

    #     Parameters
    #     ----------
    #     values : list or ndarray
    #         The values which have to be optimized.
    #     secondary : bool, optional
    #         True (default) means the secondary (storage) fluid.
    #     verbose : bool, optional
    #         Print optimization progress and result.

    #     Raises
    #     ------
    #     ValueError
    #         If the optimization does not converge.

    #     Returns
    #     -------
    #     None.

    #     """
    #     res = self.find_pinch_root(self.h_out_s, secondary, verbose)
    #     t_pinch = res[0, -1, 0]
    #     p_pinch = res[0, -1, 1]
    #     h_pinch = res[0, -1, 2]
    #     if not secondary:
    #         return t_pinch, p_pinch, h_pinch
    #     return p_pinch

    # def plot_pinch_point(self, plot_fname="", plotting=True, **kwargs):
    #     """Plot the pinch point information."""
    #     import matplotlib.pyplot as plt

    #     # Setting up plot details
    #     title = kwargs.get("title", "Pinch Point Plot")
    #     xlabel = kwargs.get("xlabel", "Enthalpy [kJ/kg]")
    #     ylabel = kwargs.get("ylabel", "Temperature [°C]")
    #     legend_loc = kwargs.get("legend_loc", "best")

    #     # Extracting necessary data
    #     working_fluid_data = self.all_states[0]
    #     storage_fluid_data = self.all_states[1]

    #     # Plotting
    #     plt.figure(figsize=(10, 6))
    #     plt.plot(
    #         working_fluid_data[:, 2],  # Enthalpy (h)
    #         working_fluid_data[:, 0],  # Temperature (T)
    #         label="Working Fluid",
    #         color="blue",
    #     )
    #     plt.plot(
    #         storage_fluid_data[:, 2],
    #         storage_fluid_data[:, 0],
    #         label="Storage Fluid",
    #         color="red",
    #     )

    #     # Adding plot details
    #     plt.title(title)
    #     plt.xlabel(xlabel)
    #     plt.ylabel(ylabel)
    #     plt.legend(loc=legend_loc)

    #     # Saving or showing the plot
    #     if plotting:
    #         plt.show()
    #     if plot_fname:
    #         plt.savefig(plot_fname)
    
    def _calculate_state_array(self, fluid, h_range):

        h_array = np.linspace(h_range[0], h_range[1], self.points)
        values = np.zeros((self.points, 2))
        values[:, 0] = h_array
        values[:, 1] = fluid.properties.pressure
        return fluid.set_state_v(values, "HP")

    def _check_temperature_consistency(self, d_tempall):
        eps_min = -1e-3
        positive = np.any(d_tempall > 0)
        negative = np.any(d_tempall < 0)
        below = True
        if self.heating < 0:
            below = False
        crossing = (positive > 0 and negative > 0)
        wrong_side = (positive > 0 and not below) or (negative > 0 and below)
        abs_dt_min = np.abs(self.dt_min)
        difference = abs_dt_min - self.d_temp_separation_min
        # print(f"Debug: abs_dt_min = {abs_dt_min}, d_temp_separation_min = {self.d_temp_separation_min}, difference = {difference}")

        if difference < eps_min:
            self.warning = 907
            self.warning_message = "Below minimum approach temperature!"
            # print(f"907: {difference}, {abs_dt_min}, {self.d_temp_separation_min}")

        elif crossing or wrong_side:
            self.warning = 999
            self.dt_mean = 1e6
            self.warning_message = "Temperatures crossing or wrong side!"
        else:
            self.warning = 0
            self.warning_message = "All o.k."

    def _plot_heat_exchanger(self, w_array, s_array):
        plot_info = self.plot_info
        # print("Plot info:", plot_info)  # Debugging-Ausgabe
        if plot_info["what"][0] == 2:
            data_w = (w_array[:, plot_info["what"][0]] - w_array[:,
                      plot_info["what"][0]].min()) * self.m_dot_w + plot_info["x-shift"][0]
            data_s = (s_array[:, plot_info["what"][0]] - s_array[:,
                      plot_info["what"][0]].min()) * self.m_dot_s + plot_info["x-shift"][1]
            plot_info["ax"].plot(data_w, w_array[:, plot_info["what"][1]],
                                 plot_info["col"][0], label=plot_info["label"][0])
            plot_info["ax"].plot(data_s, s_array[:, plot_info["what"][1]],
                                 plot_info["col"][1], label=plot_info["label"][1])
        else:
            print(
                f"H-Ex: plotting only implemented for T-H_dot [2,0]. You requested {plot_info['what']}")

    def _print_verbose(self, d_tempall):
        print(f"Min T-distance {self.dt_min}, Mean T-distance {self.dt_mean}")
        if self.heating > 0:
            print("cond", d_tempall[0], d_tempall[-1], d_tempall.min(), d_tempall.max())
        else:
            print("evap", d_tempall[0], d_tempall[-1], d_tempall.max(), d_tempall.min())

    def save_configuration_to_file(self, file_path):
        """Save the heat exchanger configuration to a file."""
        config_data = self.to_dict()
        with open(file_path, "w") as file:
            if file_path.endswith(".json"):
                json.dump(config_data, file, indent=4)
            elif file_path.endswith(".yaml"):
                yaml.dump(config_data, file)
            else:
                raise ValueError("Unsupported file format. Use .json or .yaml")

    @staticmethod
    def load_configuration_from_file(file_path):
        """Load a heat exchanger configuration from a file."""
        with open(file_path, "r") as file:
            if file_path.endswith(".json"):
                config_data = json.load(file)
            elif file_path.endswith(".yaml"):
                config_data = yaml.safe_load(file)
            else:
                raise ValueError("Unsupported file format. Use .json or .yaml")
        return config_data

    def calc(self, **kwargs):
        """Override calc method for specific calculation logic."""
        self.pinch_calc(verbose=kwargs.get("verbose", False))

    def plot(self, **kwargs):
        """Override plot method for specific plotting logic."""
        self.plot_pinch_point(plot_fname=kwargs.get(
            "plot_fname", ""), plotting=True, **kwargs)

    def save(self, filename, **kwargs):
        """Override save method for specific saving logic."""
        self.save_configuration_to_file(filename)

    def _costs(self, filename, **kwargs):
        """Override _costs method for specific cost calculation."""
        pass

def from_dict(input_hex, exchanger_name="condenser", temp_ambient=None, **kwargs):
    if not temp_ambient:
        temp_ambient = cb.CB_DEFAULTS["General"]["T_SUR"]

    working_fluid_info = input_hex.get("working_fluid")
    storage_fluid_info = None

    for key, value in input_hex.items():
        if "storage" in key.lower():
            storage_fluid_info = value
            break

    if not working_fluid_info or not storage_fluid_info:
        raise ValueError(
            "Missing working fluid or storage fluid information in input data.")

    working_fluid = cb.init_fluid(working_fluid_info["species"],
                                  working_fluid_info["fractions"],
                                  props=working_fluid_info["props"])
    storage_fluid = cb.init_fluid(storage_fluid_info["species"],
                                  storage_fluid_info["fractions"],
                                  props=storage_fluid_info["props"])

    species_mapping = input_hex[exchanger_name]["species"]
    working_fluid_states = species_mapping["working_fluid"]
    storage_fluid_states = next(
        value for key, value in species_mapping.items() if "storage" in key.lower())

    temp_out_storage = storage_fluid_info[storage_fluid_states["out"]]
    temp_in_storage = storage_fluid_info[storage_fluid_states["in"]]
    if temp_in_storage == "ambient":
        temp_in_storage = temp_ambient
    if temp_out_storage == "ambient":
        temp_out_storage = temp_ambient

    p_low_storage = storage_fluid_info["p_low"]

    state_out_storage = storage_fluid.set_state(
        [temp_out_storage, p_low_storage], "TP")
    state_in_storage = storage_fluid.set_state([temp_in_storage, p_low_storage], "TP")

    h_in_storage = state_in_storage[2]
    h_out_storage = state_out_storage[2]

    dt_min = input_hex[exchanger_name]["dt_min"]
    temp_in_working = working_fluid_info[working_fluid_states["in"]]
    temp_out_working = working_fluid_info[working_fluid_states["out"]]
    if temp_in_working == "ambient":
        temp_in_working = temp_ambient
    if temp_out_working == "ambient":
        temp_out_working = temp_ambient

    if temp_in_working > temp_in_storage:
        heating = 1
        temp_in_working += dt_min
        temp_out_working += dt_min
    else:
        heating = -1
        temp_in_working -= dt_min
        temp_out_working -= dt_min

    p_low_working = working_fluid_info["p_low"]

    state_out_working = working_fluid.set_state(
        [temp_out_working, p_low_working], "TP")
    state_in_working = working_fluid.set_state([temp_in_working, p_low_working], "TP")

    h_in_working = state_in_working[2]
    h_out_working = state_out_working[2]

    return [working_fluid, storage_fluid], input_hex[exchanger_name]["q_dot"],\
        h_out_working, h_out_storage, dt_min,\
        # input_hex[exchanger_name].get("calc_type", "const"),\
        # input_hex[exchanger_name].get("points", 50),\
        # exchanger_name,\
        # kwargs
        
        
if __name__ == "__main__":
    input_hex = {
        'condenser': {
            'model': "StaticHeatExchanger",
            "calc_type": "const",
            "species": {"working_fluid": {"in": "temp_high", "out": "temp_low"},
                        "hot_storage": {"in": "temp_low", "out": "temp_high"}},
            "dt_min": 3.,
            "q_dot": 1000.,
            "overall_u": 100,
            'area': 1,
        },
        'hot_storage': {
            "species": "Water",
            "fractions": [1.0],
            "p_low": 2e5,
            "temp_low": "ambient",
            "temp_high": 370.0,
            'props': "REFPROP",
        },
        'working_fluid': {
            "species": "Propane * Butane * Pentane * Hexane",
            "fractions": [
                7.71733787e-01,
                2.22759488e-02,
                1.78685867e-01,
                0.027304397199999997],
            "temp_low": "ambient",
            "temp_high": 420.0,
            "p_low": 1.128250708e+05,
            "p_high": 1.37548728e+06,
            "optimize": "None",
            "setting": "initial",
            'props': "REFPROP",
        },
    }

    heat_exchanger_instance = StaticHeatExchanger(
        input_hex,  "condenser")
    heat_exchanger_instance.calc()
    heat_exchanger_instance.pinch_plot()
    if heat_exchanger_instance.warning:
        print(heat_exchanger_instance.warning_message)



# -*- coding: utf-8 -*-
"""
Created on Sat Aug  3 08:40:39 2024

@author: atakan
"""
import json
import warnings
import yaml
import pandas as pd

import carbatpy as cb
from carbatpy.helpers.ser_dict import Serializable, DataAll, DataNode


class HeatExc:  # (Serializable):
    """ Base class for heat exchangers.
    
    All inputs come either from a file (json/yaml) or from a dictionary.
    
    """
    
    def __init__(self,  inputs, name=None, **kwargs):
        """
        Set the heat exchanger input parameters.

        Parameters
        ----------
        inputs : string or dictionary
            if string, it must be a valid filename (json or yaml) with the dictionary
            content. In 
        name : string, optional
            the name of the heat exchanger, the program will search for this key it in the
            inputs. The default is None.
        **kwargs : dictionary
            - plot_info : a dictionary with file, axes (from matplotlib), labels, symbols
              used for plotting (as defined in self.plot())
            DESCRIPTION.

        Raises
        ------
        NotImplementedError
            DESCRIPTION.

        Returns
        -------
        None.

        """
        if not name:
            name = "heat-exchanger"
        self.plot_info = kwargs.get("plot_info", None)
        self.output = Serializable()
        self.all_states = {}
        self.name = name
        self.warning = DataAll(value=0, message="All o.k.")
        self._defaults = DataAll()
       

        if isinstance(inputs, (dict)):
            self.inputs = inputs
        elif isinstance(inputs, (str)):
            try:
                with open(inputs, "r") as file:
                    if "json" in inputs:
                        self.inputs = json.load(file)
                    elif "yaml" in inputs:
                        self.inputs = yaml.safe_load(file)
            except Exception as e:
                print(f"Problem in heat_ex: {e}")
        else:
            raise NotImplementedError(
                f"Heat exchanger Inputs are not installed for {type(inputs), inputs}!")
        self._inp = DataAll(**self.inputs)
        
        match self._inp.process[self.name]: #noqa
            case "StaticHeatExchanger":
                pass
            case _:
                raise NotImplementedError(f"This heat exchnager is not implemented yet: {self._inp.process[self.name]}")
        
        

    def calc(self, **kwargs):
        """
        Selects the calculation method and calls the appropriate function.
        
        Depending on self.inputs.model
        """
        
                
        raise NotImplementedError("This method must be overridden in subclasses.")
        
    def plot(self, **kwargs):
        """ Plot the results into a given or a new figure.
        Dependes on self.plot_info
        """
        raise NotImplementedError("This method must be overridden in subclasses.")
        
    def save(self, filename, **kwargs):
        """Save all results into one or several files."""
        raise NotImplementedError("This method must be overridden in subclasses.")
        
    def _costs(self, filename, **kwargs):
        """ Costs are calculated via calc()."""
        raise NotImplementedError("This method must be overridden in subclasses.")


if __name__ == "__main__":

    inp = r"C:/Users/atakan/sciebo/results/test_input3.yaml"
    hex0 = HeatExc(inp, "condenser")
    
    input_hex = {
        'condenser': {
            'model': "StaticHeatExchanger",
            "calc_type": "const",
            "species": {"working_fluid": {"in": "temp_high", "out": "temp_low"},
                        "hot_storage": {"in": "temp_low", "out": "temp_high"}},
            "dt_min": 3.,
            "q_dot": 1000.,
            "overall_u": 100,
            'area': 1,
        },
        'hot_storage': {
            "species": "Water",
            "fractions": [1.0],
            "p_low": 2e5,
            "temp_low": "ambient",
            "temp_high": 370.0,
            'props': "REFPROP",
        },
        'working_fluid': {
            "species": "Propane * Butane * Pentane * Hexane",
            "fractions": [
                7.71733787e-01,
                2.22759488e-02,
                1.78685867e-01,
                0.027304397199999997],
            "temp_low": "ambient",
            "temp_high": 420.0,
            "p_low": 1.128250708e+05,
            "p_high": 1.37548728e+06,
            "optimize": "None",
            "setting": "initial",
            'props': "REFPROP",
        },
    }



# -*- coding: utf-8 -*-
"""
Some constants for carbatpy usage are set here

* The two (!) Refprop installation directories (with dll, fluids etc.)
* he directory, where the results shall be written.

Created on Sun Nov  5 16:01:02 2023

@author: atakan
"""

import os
import copy
verbose = False
# global _T_SURROUNDING, _P_SURROUNDING, _RESULTS_DIR, _CARBATPY_BASE_DIR
_T_SURROUNDING = 288.15 # K
_P_SURROUNDING =1.013e5  # Pa
TREND = {"TREND_INSTALLED":True,
          "USE_TREND":False,
          "TREND_DLL":"",
          'TREND_PATH':""}

script_path = os.path.abspath(__file__)
script_dir = os.path.dirname(script_path)
parent_dir = os.path.dirname(script_dir)
# Standard directory of the installation
grandparent_dir = os.path.dirname(parent_dir)
if verbose:
    print(f"Standard Carbatpy base directory: {grandparent_dir}")

    print(f"Directory of the main source code: {script_dir}")
directory = os.getcwd()
try:
    _CARBATPY_BASE_DIR = os.environ["CARBATPY_BASE_DIR"]
    print(f"You are using this custom set ('CARBATPY_BASE_DIR') base directory for carbatpy: {script_dir}")
except:
    if script_dir.count('carbatpy') >0:
        _CARBATPY_BASE_DIR = script_dir
    else:
        _CARBATPY_BASE_DIR = grandparent_dir
if verbose: print(script_dir.count('carbatpy') >0)
# The two installations of REFPROP , one for the working fluid
# and one for the secondary fluid. With only one installation, the instances
# mix up ...
os.environ['RPPREFIX'] = r'C:/Program Files (x86)/REFPROP'
os.environ['RPPREFIXs'] = r'C:/Program Files (x86)/REFPROP/secondCopyREFPROP'
if TREND["TREND_INSTALLED"]:
    try:
        TREND["TREND_DLL"] = os.environ['TREND_DLL']
        TREND["TREND_PATH"] = os.environ['TREND_PATH']
    except:
        print("Trend not found! Check the environmentvariable TREND_DLL, TREND_PATH")
        TREND_DLL = ""
        TREND_PATH = ''
        TREND["TREND_INSTALLED"] = False

try:
    _RESULTS_DIR = os.environ['CARBATPY_RES_DIR']
except:
    try:
        _RESULTS_DIR = os.environ['TEMP']
    except:
        try:

            _RESULTS_DIR = directory + r"\\tests\\test_files"
        except Exception as no_file:
            print("Please set the envirionment variable: CARBATPY_RES_DIR !", no_file)

# Default values:-------------------------------------------------
fl_properties_names = ("Temperature",
                       "Pressure",
                       "spec_Enthalpy",
                        "spec_Volume",
                        "spec_Entropy",
                        "quality_mass",
                        "spec_internal_Energy",
                        "viscosity",
                        "thermal_conductivity",
                        "Prandtl_number",
                        "isobaric_heat_capacity",
                        "speed_of_sound",
                        "molecular_mass")

fl_pr_variable_names = {"temperature":'T',
                        'pressure': 'p',
                        'enthalpy':'h',
                        "sp_volume": "v",
                        'entropy': 's',
                        'quality': 'q',
                        'int_energy': 'u',
                        'viscosity': "eta",
                        'thermal_conductivity': "k",
                        'prandtl': "Pr",
                        "cp":"cp",
                        'speed_of_sound': "w_s",
                        'molecular_mass': "M",
                        }

THERMO_STRING = "T;P;H;V;S;QMASS;E"
TRANS_STRING = THERMO_STRING + ";VIS;TCX;PRANDTL;CP;W;M"
THERMO_TREND = request = ["T", "P", "H", "D",
                           "S", "QEOS", "U"]  # careful density not volume
TRANS_TREND = copy.copy(THERMO_TREND)
TRANS_TREND.extend(["ETA", "TCX", "CP","WS"]) # no molecular mass

                           
TRANS_TREND_MIX = copy.copy(THERMO_TREND)
TRANS_TREND_MIX.extend(["ETA_ECS", "TCX_ECS", "CP","WS"]) # no molecular mass

CB_FLUID_DEFAULT = {"PROPS": "REFPROP",  # choice: "TREND" or "REFPROP"
                    "DLL_SELECT": "2dll",  # choice: "2dll" or "dll"
                    "UNITS": "MASS BASE SI",
                    "Property_Names":fl_properties_names,
                    "Property_Names_Short":fl_pr_variable_names,
                    "THERMO_STRING":THERMO_STRING,
                    "TRANS_STRING":TRANS_STRING ,
                    "THERMO_TREND": THERMO_TREND,
                    "TRANS_TREND": TRANS_TREND,
                    "TREND":TREND,
                    "RPPREFIX":os.environ['RPPREFIX'],
                    "TRANS_TREND_MIX": TRANS_TREND_MIX,



    }
# This dictionary will be generally imported with carbatpy
CB_DEFAULTS={
    "Fluid_Defaults": CB_FLUID_DEFAULT,
    "General":{"T_SUR":_T_SURROUNDING,
               "P_SUR": _P_SURROUNDING,
               "RES_DIR": _RESULTS_DIR,
               "CB_DIR": _CARBATPY_BASE_DIR,
               "CB_DATA":_CARBATPY_BASE_DIR+"\\data"

               }
    }


# -*- coding: utf-8 -*-
"""
Base class for thermodynamic cycles
Created on Tue Jul 30 13:26:28 2024

@author: atakan
"""
import copy
import numpy as np
import carbatpy as cb
from carbatpy.helpers.ser_dict import Serializable

class Cycle(Serializable):
    def __init__(self, fixed_points, components=None, **kwargs):
        self.fixed_points = fixed_points
        self.components = components if components else {}
        self.kwargs =kwargs
        self.all_states = []
        self.m_dots = []
        self.warning = []
        self.evaluation = self.init_evaluation()

        self.specific_init()
        
    

    
    
    def set_fl_state(self, fluid_values, fixed_points, **kwargs):
        # Implementierung bleibt spezifisch für Unterklassen
        raise NotImplementedError("This method must be overridden in subclasses.")

    def get_fp_value(self, key):
        # Keine Unterscheidung nach "common", direkt aus fixed_points entnehmen
        value = self.fixed_points.get(key)
        if value is None:
            raise KeyError(f"{key} not found in the input data.")
        return value
    
    

    def calc_cycle(self):
        raise NotImplementedError("This method must be overridden in subclasses.")

    
    """"    
    def set_fl_state(self, fluid_values, fixed_points, **kwargs):
        # Initialize fluid
        fluid_act = cb.init_fluid(*fluid_values["common"]["fluids_all"]["value"]["WORKING"][1:])
        fl_type = fluid_values["common"]["fluids_all"]["value"]["WORKING"][0]

        # Set common state values
        d_t_min = fluid_values["common"]["D_T_MIN"]["value"]
        d_t_super = fluid_values["common"]["D_T_SUPER"]["value"]

        # Check if it's a storage type
        if "STORAGE" in fl_type:
            fluid_act, fixed_points = self._set_storage_fl_state(fl_type, fluid_act, fixed_points)
        else:
            fluid_act, fixed_points = self._set_working_fl_state(fl_type, fluid_act, fixed_points, fluid_values, d_t_min, d_t_super)

        return fluid_act, fixed_points
    
    def get_fp_value(self, key):
        # Attempt to fetch value from 'common' section first
        value = self.fixed_points.get("common", {}).get(key, {}).get("value")
        if value is None:
            # Attempt to fetch value from specific section (based on the instance type)
            if isinstance(self, HeatPump):
                value = self.fixed_points.get("heat_pump", {}).get(key)
            elif isinstance(self, ORC):
                value = self.fixed_points.get("orc", {}).get(key)
        if value is None:
            raise KeyError(f"{key} not found in the input data.")
        return value
    
    def init_fluids(self):
        for fll, ii in self.fluids_all.items():
            # Setze den Zustand der Fluide basierend auf den Fixpunkten
            self.fluids.append(self.set_fl_state([fll, *ii], self.fixed_points)[0])

    def _set_storage_fl_state(self, fl_type, fluid_act, fixed_points):
        p_in, t_in, t_out = self._extract_storage_values(fixed_points, fl_type)
        state_out = fluid_act.set_state([t_out, p_in], "TP")
        state_in = fluid_act.set_state([t_in, p_in], "TP")

        if "HOT" in fl_type:
            fixed_points['h_h_out_sec'] = state_out[2]
        elif "COLD" in fl_type:
            fixed_points['h_l_out_cold'] = state_out[2]

        return fluid_act, fixed_points

    def _set_storage_fl_state(self, fl_type, fluid_act, fixed_points):
        p_in, t_in, t_out = self._extract_storage_values(fixed_points, fl_type)
        state_out = fluid_act.set_state([t_out, p_in], "TP")
        state_in = fluid_act.set_state([t_in, p_in], "TP")

        if "HOT" in fl_type:
            fixed_points['h_h_out_sec'] = state_out[2]
        elif "COLD" in fl_type:
            fixed_points['h_l_out_cold'] = state_out[2]

        return fluid_act, fixed_points

    def _set_working_fl_state(self, fl_type, fluid_act, fixed_points, fluid_values, d_t_min, d_t_super):
        w_p_select = fluid_values.get("P_WORKING", {"setting": "initial"})
        if w_p_select["setting"] == "auto":
            fluid_act, fixed_points = self._auto_set_pressure(fluid_act, fixed_points, d_t_min, d_t_super)
        else:
            fluid_act, fixed_points = self._manual_set_pressure(fluid_act, fixed_points, w_p_select, d_t_min, d_t_super)

        return fluid_act, fixed_points

    def _auto_set_pressure(self, fluid_act, fixed_points, d_t_min, d_t_super):
        state_in_cond = fluid_act.set_state([fixed_points['T_OUT_STORAGE_HOT'], 1.], "TQ")
        p_high = state_in_cond[1]

        state_out_cond = fluid_act.set_state([fixed_points['T_IN_STORAGE_HOT'] + d_t_min, p_high], "TP")
        T_IN = fixed_points['T_IN_STORAGE_HOT'] - d_t_min
        state_satv_evap = fluid_act.set_state([T_IN - d_t_super, 1.], "TQ")
        p_low = state_satv_evap[1]

        state_out_evap = fluid_act.set_state([p_low, T_IN], "PT")

        fixed_points["p_low"] = p_low
        fixed_points["p_high"] = p_high
        fixed_points['h_h_out_w'] = state_out_cond[2]
        fixed_points['h_l_out_w'] = state_out_evap[2]

        return fluid_act, fixed_points

    def _manual_set_pressure(self, fluid_act, fixed_points, w_p_select, d_t_min, d_t_super):
        p_high = w_p_select.get("p_high", 0)
        p_low = w_p_select.get("p_low", 0)

        state_in_cond = fluid_act.set_state([fixed_points['T_OUT_STORAGE_HOT'], p_high], "TP")
        state_out_cond = fluid_act.set_state([fixed_points['T_IN_STORAGE_HOT'] + d_t_min, p_high], "TP")

        T_IN = fixed_points['T_IN_STORAGE_HOT'] - d_t_min
        state_satv_evap = fluid_act.set_state([T_IN - d_t_super, p_low], "TP")
        state_out_evap = fluid_act.set_state([p_low, T_IN], "PT")

        fixed_points["p_low"] = p_low
        fixed_points["p_high"] = p_high
        fixed_points['h_h_out_w'] = state_out_cond[2]
        fixed_points['h_l_out_w'] = state_out_evap[2]

        return fluid_act, fixed_points

    def _extract_storage_values(self, fixed_points, fl_type):
        # Extracts storage values for specific fluid type (hot/cold)
        if "HOT" in fl_type:
            p_in = fixed_points.get('P_IN_STORAGE_HOT', 0)
            t_in = fixed_points.get('T_IN_STORAGE_HOT', 0)
            t_out = fixed_points.get('T_OUT_STORAGE_HOT', 0)
        elif "COLD" in fl_type:
            p_in = fixed_points.get('P_IN_STORAGE_COLD', 0)
            t_in = fixed_points.get('T_IN_STORAGE_COLD', 0)
            t_out = fixed_points.get('T_OUT_STORAGE_COLD', 0)
        else:
            raise ValueError(f"Unknown storage type: {fl_type}")
        return p_in, t_in, t_out
    
     """   
    def calculate_hex(self, params, plot_info={}, verbose_=True):
        """
        General method for calculating a heat exchanger.

        params: dict containing the following keys:
            - fluid_numbers: List[int]
            - h_dot_min: float
            - h_out_w: float
            - h_limit_s: float (optional)
            - d_temp_separation_min: float (optional)
            - calc_type: str (optional)
            - pinch: str (optional)
            - name: str (optional)
        """
        fluid_numbers = params["fluid_numbers"]
        h_dot_min = params["h_dot_min"]
        h_out_w = params["h_out_w"]
        h_limit_s = params.get("h_limit_s", np.nan)
        d_temp_separation_min = params.get("d_temp_separation_min", 0.5)
        calc_type = params.get("calc_type", "const")
        pinch = params.get("pinch", "")
        name = params.get("name", "evaporator")

        w_actual, s_actual = [self.fluids[fluid_numbers[0]],
                              self.fluids[fluid_numbers[1]]]
        hex_act = cb.hex_th.StaticHeatExchanger(
            [w_actual, s_actual],
            h_dot_min, h_out_w, h_limit_s,
            d_temp_separation_min=d_temp_separation_min,
            name=name
        )

        if pinch == "":
            h_s_out = hex_act.pinch_calc(plot_info)
        elif pinch == "secondary":
            h_s_out = hex_act.find_pinch()
        elif pinch == "working":
            h_s_out = hex_act.find_pinch(False)
        else:
            print("calculate_hex: This option is not available")

        if hex_act.warning > 0:
            print("Heat-Exchanger problem: ", hex_act.warning, hex_act.warning_message)
            self.warning.append([hex_act.name, hex_act.warning, hex_act.warning_message])
        
        m_dot_s, d_tempall, wf_states, sf_states = hex_act.pinch_plot(plotting=verbose_)
        d_temp_mean = hex_act.dt_mean
        d_temp_min = hex_act.dt_min

        self.all_states.append(wf_states)
        self.all_states.append(sf_states)
        self.m_dots.append(hex_act.m_dot_w)
        self.m_dots.append(m_dot_s)

        w_after_hex = self.fluids[0].set_state([wf_states[0, 1], wf_states[0, 2]], "PH")
        fluids_states_in_out = np.array([[wf_states[0], wf_states[-1]], [sf_states[0], sf_states[-1]]])
        fluids_states_in_out_ex = fluids_states_in_out
        if hex_act.heating < 1:
            fluids_states_in_out_ex = np.flip(fluids_states_in_out, axis=1)

        mass_flow_rates = np.array([hex_act.m_dot_w, hex_act.m_dot_s])
        ex_loss_rate = cb.exlo.exergy_loss_flow(fluids_states_in_out_ex, mass_flow_rates)
        
        self.evaluation[name] = {
            "states-in-out": fluids_states_in_out,
            "mass-flow-rates": mass_flow_rates,
            "exergy-loss-rate": ex_loss_rate,
            "dT-min": d_temp_min,
            "dT-mean": d_temp_mean
        }
        self.components[name] = hex_act.all
        
        if verbose_:
            print(f"After heat exchanger: {name}")
            w_actual.print_state()
        
        return w_after_hex

          
    

    

    def plot(self, f_name, fig_ax=[]):
        # Gemeinsame Plot-Methode
        pass

    # Weitere gemeinsame Methoden und Attribute


class HeatPump(Cycle):
    def __init__(self, components=None, fixed_points=None, fluids=None, **kwargs):
        super().__init__(components=components, fixed_points=fixed_points, **kwargs)
        # Additional initializations specific to HeatPump
        self.fluids_all = fixed_points.get("fluids_all", {})
        self.fluids = []
        super().__init__(fixed_points, components)
        self.init_fluids()
        
    def init_evaluation(self):
       # Definition der spezifischen Evaluationsattribute für HeatPump
       return {
           "Q_DOT_MIN": self.fixed_points["common"]["Q_DOT_MIN"],
           "Power": 0.0,
           "T_OUT_STORAGE_HOT": self.fixed_points["common"]["T_OUT_STORAGE_HOT"],
           "T_OUT_STORAGE_COLD": self.fixed_points["common"]["T_OUT_STORAGE_COLD"],
           "exergy_loss_rate": 0
       }

    def specific_init(self):
        # Zusätzliche Initialisierungen für HeatPump
        pass
    
    def _set_working_fl_state(self, fl_type, fluid_act, fixed_points):
        # If HeatPump has a specific way of setting working fluid states
        return super()._set_working_fl_state(fl_type, fluid_act, fixed_points)
    
    

    def calc_cycle(self, plot_info={}, variations=["", ""], verbose=False):
        """
        Calculates a simple compression heat pump cycle.
        """
        state_in = copy.copy(self.fluids[0])
        w_actual = self.fluids[0]
        if verbose:
            print("Start:")
            w_actual.print_state()

        # Compressor
        compressor = cb.FlowDevice(
            self.fluids[0], self.fixed_points["p_high"], 1.0,
            device_type="machine", name="compressor",
            calc_type="const_eta", calc_parameters={"eta_s": self.fixed_points["ETA_S_C"]},
            plot_info=plot_info
        )
        w_after_compressor, work_c, _power = compressor.state_w_p()
        self.components[compressor.name] = compressor.all
        
        if verbose:
            print("After the compressor")
            w_actual.print_state()

        # Condenser
        params = {
            "fluid_numbers": [0, 1],
            "h_dot_min": self.fixed_points["Q_DOT_MIN"],
            "h_out_w": self.fixed_points["h_h_out_w"],
            "h_limit_s": self.fixed_points["h_h_out_sec"],
            "d_temp_separation_min": self.fixed_points['D_T_MIN'],
            "calc_type": variations[0],
            "name": "condenser"
        }
        w_after_condenser = self.calculate_hex(params, plot_info=plot_info, verbose_=verbose)

        power = work_c * self.m_dots[params["fluid_numbers"][0]]
        heat_flow_evap = self.fixed_points["Q_DOT_MIN"] - power
        if verbose:
            print(f"Power: {power} W, Evaporator heat: {heat_flow_evap} W\n")

        # Throttle
        throttle = cb.FlowDevice(
            w_actual, self.fixed_points["p_low"], self.m_dots[params["fluid_numbers"][0]],
            device_type="throttle", name="throttle-A", calc_type="const_h", calc_parameters={},
            plot_info=plot_info
        )
        w_after_throttle = throttle.output["state_out"]
        self.components[throttle.name] = throttle.all

        if verbose:
            print("After the throttle:")
            print(w_actual.print_state())

        w_act = w_actual.set_state([w_after_throttle[1], w_after_throttle[2]], "PH")
        if self.fixed_points["T_OUT_STORAGE_COLD"] < w_act[0] + self.fixed_points['D_T_MIN']:
            self.fixed_points["T_OUT_STORAGE_COLD"] = w_act[0] + self.fixed_points['D_T_MIN']

        # Evaporator
        params = {
            "fluid_numbers": [0, 2],
            "h_dot_min": heat_flow_evap,
            "h_out_w": self.fixed_points["h_l_out_w"],
            "h_limit_s": self.fixed_points["h_l_out_cold"],
            "d_temp_separation_min": self.fixed_points['D_T_MIN'],
            "calc_type": variations[0],
            "name": "evaporator"
        }
        w_after_evap = self.calculate_hex(params, plot_info=plot_info, verbose_=verbose)

        q_evap = (w_after_evap[2] - w_after_throttle[2]) * self.m_dots[params["fluid_numbers"][0]]
        if heat_flow_evap - q_evap > 1:
            print(f"Heat pump cycle not in equilibrium: {heat_flow_evap}, {q_evap}")

        if verbose:
            print(f"Evaporator heat flow rate: {q_evap}, Power: {power}")
            w_actual.print_state()

        # Calculate exergy losses
        sec_fluids_states_in_out = np.array([
            self.evaluation["condenser"]["states-in-out"][-1],
            self.evaluation["evaporator"]["states-in-out"][-1]
        ])
        mass_flow_rates = np.array([
            self.evaluation["condenser"]["mass-flow-rates"][-1],
            self.evaluation["evaporator"]["mass-flow-rates"][-1]
        ])
        ex_loss_rate = cb.exlo.exergy_loss_flow(sec_fluids_states_in_out, mass_flow_rates)

        cop = self.fixed_points["Q_DOT_MIN"] / power
        self.evaluation.update({
            "Q_DOT_MIN": self.fixed_points["Q_DOT_MIN"],
            "Power": power,
            "p_high": self.fixed_points["p_high"],
            "p_low": self.fixed_points["p_low"],
            "exergy_loss_rate": ex_loss_rate,
            "sec_fluid_states_in_out": sec_fluids_states_in_out,
            "sec_fluid_m_dots": mass_flow_rates,
            "COP": cop
        })
        return cop

    # def calculate_hex(self, fluid_numbers, h_dot_min_, h_out_w_, h_limit_s_=float('nan'), points_=50, d_temp_separation_min_=0.5, calc_type_='const', pinch='', name_='evaporator', plot_info={}, verbose_=True):
    #     # Weitere spezifische Methode
    #     pass


class OrganicRankineCycle(Cycle):
    def __init__(self, components=None, fixed_points=None, fluids=None, **kwargs):
        super().__init__(components=components, fixed_points=fixed_points, fluids=fluids, **kwargs)
        # Additional initializations specific to ORC
        
        self.fluids_all = fixed_points.get("fluids_all", {})
        self.fluids = []
        super().__init__(fixed_points, components)
        self.init_fluids()
        
    def init_evaluation(self):
        # Definition der spezifischen Evaluationsattribute für ORC
        return {
            "Q_DOT_MIN": self.fixed_points["Q_DOT_MIN"],
            "Power-Net": 0.0,
            "T_hh": self.fixed_points["T_OUT_STORAGE_HOT"], #reversed reative to heat pump
            "T_ll": self.fixed_points["T_OUT_STORAGE_COLD"],
            "eta_pc": 0
        }

    def specific_init(self):
        # Zusätzliche Initialisierungen für ORC
        pass
    
    def _set_working_fl_state(self, fl_type, fluid_act, fixed_points):
        # If ORC has a specific way of setting working fluid states
        return super()._set_working_fl_state(fl_type, fluid_act, fixed_points)
    
    def calc_cycle(self, verbose=False):
        """
        Calculates a simple Organic Rankine Cycle (ORC).
        """
        state_in = copy.copy(self.fluids[0])
        w_actual = self.fluids[0]
        if verbose:
            print("Start:")
            w_actual.print_state()

        # Pump
        pump = cb.FlowDevice(
            self.fluids[0], self.fixed_points["p_high"], 1.0,
            device_type="pump", name="pump",
            calc_type="const_eta", calc_parameters={"eta_s": self.fixed_points["eta_s_p"]}
        )
        w_after_pump, work_pump, _power = pump.state_w_p()
        m_dot_w = self.fixed_points["Q_DOT_MIN"] / (self.fixed_points["h_out_w"] - w_after_pump[2])
        self.components[pump.name] = pump.all

        if verbose:
            print("After the pump")
            w_actual.print_state()

        # Evaporator
        params = {
            "fluid_numbers": [0, 1],
            "h_dot_min": self.fixed_points["Q_DOT_MIN"],
            "h_out_w": self.fixed_points["h_out_w"],
            "d_temp_separation_min": self.fixed_points["D_T_MIN"],
            "calc_type": self.fixed_points["Q_DOT_MIN"],
            "name": "evaporator"
        }
        w_after_evaporator = self.calculate_hex(params, verbose_=verbose)

        # Expander
        expander = cb.FlowDevice(
            self.fluids[0], self.fixed_points["p_low"], m_dot_w,
            device_type="machine", name="expander",
            calc_type="const_eta", calc_parameters={"eta_s": self.fixed_points["eta_s_e"]}
        )
        w_after_expander, work_expander, _power = expander.state_w_p()
        self.components[expander.name] = expander.all

        if verbose:
            print("After the expander")
            w_actual.print_state()

        # Condenser
        params = {
            "fluid_numbers": [0, 2],
            "h_dot_min": self.fixed_points["Q_DOT_MIN"],
            "h_out_w": self.fixed_points["h_out_c"],
            "d_temp_separation_min": self.fixed_points["D_T_MIN"],
            "calc_type": self.fixed_points["Q_DOT_MIN"],
            "name": "condenser"
        }
        w_after_condenser = self.calculate_hex(params, verbose_=verbose)

        # Net work and efficiency
        work_net = work_expander - work_pump
        eta = work_net / self.fixed_points["Q_DOT_MIN"]

        self.evaluation.update({
            "NetWork": work_net,
            "p_high": self.fixed_points["p_high"],
            "p_low": self.fixed_points["p_low"],
            "Efficiency": eta
        })
        return eta


import json
"""abrik-Muster für die Erzeugung von Kreisprozessen

Ein Fabrik-Muster kann genutzt werden, um basierend auf einer Konfigurationsdatei oder Parameterinstanz den entsprechenden Kreisprozess zu erstellen"""


class CycleFactory:
    @staticmethod
    def create_cycle(config_file):
        with open(config_file, 'r') as f:
            config = json.load(f)
        
        cycle_type = config.get('type')
        fixed_points = config.get('fixed_points')
        components = config.get('components', [])

        if cycle_type == 'HeatPump':
            return HeatPump(fixed_points, components)
        elif cycle_type == 'OrganicRankineCycle':
            fluids = config.get('fluids')
            return OrganicRankineCycle(fluids, fixed_points, components)
        else:
            raise ValueError(f"Unknown cycle type: {cycle_type}")


""" config.json
{
    "type": "HeatPump",
    "fixed_points": [300, 350],
    "components": ["compressor", "condenser"],
    "plot_info": {}
}

"""

if __name__ =="__main__":
    import json
    file_n = cb.CB_DEFAULTS["General"]["CB_DATA"]+"\\input_hp_orc_unified.json"
    with open(file_n, "r") as file:
        inp_hp = json.load(file)
        
    file_n = cb.CB_DEFAULTS["General"]["CB_DATA"]+"\\orc_input.json"
    with open(file_n, "r") as file:
        inp_orc = json.load(file)
        
    hp_act =HeatPump(fixed_points=inp_hp)
    hp_act.calc_cycle()
    
    orc_act =OrganicRankineCycle(fixed_points=inp_hp)
    orc_act.calc_cycle()


# -*- coding: utf-8 -*-
"""
Created on Wed Nov 27 13:12:35 2024

@author: atakan
"""

import numpy as np
import copy

class YourClass:
    def __init__(self, name, inputs, warning, logger, _in_def, _out_def, _fixed):
        self.name = name
        self.inputs = inputs
        self.warning = warning
        self.logger = logger
        self._in_def = _in_def
        self._out_def = _out_def
        self._fixed = _fixed
        self.output = {}
        self.all_data = {}
        self._q_dot = None
        self._heating = -1

    def _storage_name(self, fluid_names):
        # Implement this method based on your logic
        pass

    def _calculate_state_array(self, w_out, states):
        # Implement this method based on your logic
        pass

    def _check_temperature_consistency(self, d_tempall):
        # Implement this method based on your logic
        pass

    def calculate(self, in_states=None, out_states=None, run_param=None, **kwargs):
        self.warning.__dict__.update(value=0, message="All o.k.")
        verbose = kwargs.get("verbose", False)
        if verbose:
            self.logger.info(f"Calculating {self.name}")

        self._run_param = run_param

        if not in_states:
            in_states = self._in_def
            self.warning.__dict__.update(value=2, message="Default values used for I/O")
        if not out_states:
            out_states = self._out_def
            self.warning.__dict__.update(value=2, message="Default values used for I/O")

        fluid_names = list(self.inputs["act_fluids"].keys())
        stn = self._storage_name(fluid_names)
        self.output.update({"m_dot": {}, "state_in": {}, "state_out": {}})
        self.all_data = {}
        w_out = {}
        self._heating = -1

        match self._fixed:
            case 'q_dot':
                # Case 2: Q, h_w_a, and h_s_a are given
                Q = self._q_dot
                h_w_e = in_states["working_fluid"][2]
                h_w_a = out_states["working_fluid"][2]
                h_s_e = in_states[stn][2]
                h_s_a = out_states[stn][2]

                m_w = Q / (h_w_e - h_w_a)
                m_s = Q / (h_s_a - h_s_e)

                self.output['m_dot']["working_fluid"] = m_w
                self.output['m_dot'][stn] = m_s

            case 'm_dot_w':
                # Case 3: m_w, h_w_a, and h_s_a are given
                m_w = run_param["m_dot"]["working_fluid"]
                h_w_e = in_states["working_fluid"][2]
                h_w_a = out_states["working_fluid"][2]
                h_s_e = in_states[stn][2]
                h_s_a = out_states[stn][2]

                Q = m_w * (h_w_e - h_w_a)
                m_s = Q / (h_s_a - h_s_e)

                self.output['m_dot']["working_fluid"] = m_w
                self.output['m_dot'][stn] = m_s
                self._q_dot = Q

            case 'q_m_dot_w':
                # Case 1: Q, m_w, and h_s_a are given
                Q = self._q_dot
                m_w = run_param["m_dot"]["working_fluid"]
                h_w_e = in_states["working_fluid"][2]
                h_s_e = in_states[stn][2]
                h_s_a = out_states[stn][2]

                h_w_a = h_w_e - Q / m_w
                m_s = Q / (h_s_a - h_s_e)

                self.output['m_dot']["working_fluid"] = m_w
                self.output['m_dot'][stn] = m_s
                out_states["working_fluid"][2] = h_w_a

            case _:
                raise NotImplementedError(f"StaticHeatExchanger: {self.inputs['parameters']['fixed']} not implemented!")

        for fln in ["working_fluid", stn]:
            w_in = copy.copy(self.inputs['act_fluids'][fln])
            self.output['state_in'][fln] = w_in.set_state([in_states[fln][1], in_states[fln][2]], "PH")
            self.output['state_out'][fln] = w_in.set_state([out_states[fln][1], out_states[fln][2]], "PH")
            w_out[fln] = copy.copy(w_in)
            # all states set now

        for fln in ["working_fluid", stn]:
            if fln == "working_fluid":  # if mdot for working fluid is fixed
                self._q_dot = (self.output['state_in'][fln][2] - self.output['state_out'][fln][2]) * self.output['m_dot'][fln]
                self.output['m_dot'][stn] = np.abs(self._q_dot / (self.output['state_in'][stn][2] - self.output['state_out'][stn][2]))
            elif self._fixed == "q_dot":  # q_dot is fixed
                if 'm_dot' in run_param:
                    if fln in run_param['m_dot']:
                        self.output['m_dot'][fln] = run_param['m_dot'][fln]
                    else:
                        self.output['m_dot'][fln] = np.abs(self._q_dot / (self.output['state_in'][fln][2] - self.output['state_out'][fln][2]))
                else:
                    raise NotImplementedError(f"HeatExch, this fixed value is not implemented {self._fixed}")

            if fln == "working_fluid":
                temp_w_in = self.output['state_in'][fln][0]
                self.all_data[fln] = self._calculate_state_array(w_out[fln], [self.output['state_in'][fln], self.output['state_out'][fln]])
            else:
                temp_s_in = self.output['state_in'][fln][0]
                self.all_data[fln] = self._calculate_state_array(w_out[fln], [self.output['state_out'][fln], self.output['state_in'][fln]])

            if temp_w_in > temp_s_in:
                self._heating = 1

        d_tempall = self.all_data["working_fluid"][:, 0] - self.all_data[stn][:, 0]
        self.dt_mean, self.dt_min, self.dt_max = d_tempall.mean(), np.abs(d_tempall).min(), np.abs(d_tempall).max()

        self._check_temperature_consistency(d_tempall)

        self.output.update({"dt_mean": self.dt_mean, "dt_min": self.dt_min, "dt_max": self.dt_max, "q_dot": -self._q_dot})
"""
# Example usage:
# Initialize your class with appropriate parameters
# your_instance = YourClass(name, inputs, warning, logger, _in_def, _out_def, _fixed)
# your_instance.calculate(in_states, out_states, run_param)
```

### Explanation:
1. **Case 1 (`q_m_dot_w`)**: If `Q`, `m_w`, and `h_s_a` are given, we calculate `h_w_a` and `m_s`.
2. **Case 2 (`q_dot`)**: If `Q`, `h_w_a`, and `h_s_a` are given, we calculate `m_w` and `m_s`.
3. **Case 3 (`m_dot_w`)**: If `m_w`, `h_w_a`, and `h_s_a` are given, we calculate `Q` and `m_s`.

The `match` statement is used to handle the different cases based on the value of `self._fixed`. The calculations are performed accordingly, and the results are stored in the `self.output` dictionary.

You can adjust the example usage to fit your specific input values and test the function accordingly.

"""


# -*- coding: utf-8 -*-
"""
Class for the input of the heat pump

values are stored in a dictionarypart of carbatpy
Created on Tue Jul 23 16:48:28 2024

@author: atakan
"""

import json
import yaml

class HpVal:
    DEFAULT_DIR =cb.CB_DEFAULTS["General"]["CB_DATA"]
    DEFAULT_FILE=DEFAULT_DIR+"\\hp-input-dictvariables"
    def __init__(self, variables_dict=None):
        if variables_dict:
            for key, value in variables_dict.items():
                setattr(self, key, value)
    
    def to_dict(self):
        return {key: getattr(self, key) for key in self.__dict__}

    def save_to_json(self, file_path=DEFAULT_FILE+"_act.json"):
        with open(file_path, 'w') as json_file:
            json.dump(self.to_dict(), json_file, indent=4)

    @classmethod
    def load_from_json(cls, file_path=DEFAULT_FILE+".json"):
        with open(file_path, 'r') as json_file:
            variables_dict = json.load(json_file)
            return cls(variables_dict)

    def save_to_yaml(self, file_path=DEFAULT_FILE+"_act.yaml"):
        with open(file_path, 'w') as yaml_file:
            yaml.dump(self.to_dict(), yaml_file, default_flow_style=False)

    @classmethod
    def load_from_yaml(cls, file_path=DEFAULT_FILE+".yaml"):
        with open(file_path, 'r') as yaml_file:
            variables_dict = yaml.safe_load(yaml_file)
            return cls(variables_dict)


if __name__ == "__main__":
    # Beispielhafte Zuweisung der Variablen und Parameter
    variables_dict = {
        'FLUID': "Propane * Butane * Pentane * Hexane",
        'comp': [0.75, 0.05, 0.15, 0.05],
        'FLS': "Water",
        'FLCOLD': "Methanol",
        'ETA_S': 0.7,
        'STORAGE_T_IN': 298.15,  # Beispielwert für cb.T_SURROUNDING
        'COLD_STORAGE_T_IN': 298.15,
        'STORAGE_T_OUT': 363.0,  # 395.0
        'COLD_STORAGE_T_OUT': 250.15,
        'STORAGE_P_IN': 5e5,
        'COLD_STORAGE_P_IN': 5e5,
        'Q_DOT_MIN': 1e3,  # heat flow rate (W)
        'D_T_SUPER': 5,  # super heating of working fluid
        'D_T_MIN': 4.0  # minimum approach temperature (pinch point)
    }
    
    # Erstellen einer Instanz der Klasse HpVal
    hp_val = HpVal(variables_dict)
    
    # Speichern des Dictionaries in einer JSON-Datei
    hp_val.save_to_json('variables.json')
    
    # Einlesen des Dictionaries aus einer JSON-Datei und Erstellen einer neuen Instanz von HpVal
    hp_val_from_json = HpVal.load_from_json('variables.json')
    
    # Ausgabe, um zu prüfen, ob die Variablen korrekt zugewiesen wurden
    print("JSON:")
    print(hp_val_from_json.FLUID)
    print(hp_val_from_json.comp)
    print(hp_val_from_json.FLS)
    print(hp_val_from_json.FLCOLD)
    print(hp_val_from_json.ETA_S)
    print(hp_val_from_json.STORAGE_T_IN)
    print(hp_val_from_json.COLD_STORAGE_T_IN)
    print(hp_val_from_json.STORAGE_T_OUT)
    print(hp_val_from_json.COLD_STORAGE_T_OUT)
    print(hp_val_from_json.STORAGE_P_IN)
    print(hp_val_from_json.COLD_STORAGE_P_IN)
    print(hp_val_from_json.Q_DOT_MIN)
    print(hp_val_from_json.D_T_SUPER)
    print(hp_val_from_json.D_T_MIN)
    
    # Speichern des Dictionaries in einer YAML-Datei
    hp_val.save_to_yaml('variables.yaml')
    
    # Einlesen des Dictionaries aus einer YAML-Datei und Erstellen einer neuen Instanz von HpVal
    hp_val_from_yaml = HpVal.load_from_yaml('variables.yaml')
    
    # Ausgabe, um zu prüfen, ob die Variablen korrekt zugewiesen wurden
    print("\nYAML:")
    print(hp_val_from_yaml.FLUID)
    print(hp_val_from_yaml.comp)
    print(hp_val_from_yaml.FLS)
    print(hp_val_from_yaml.FLCOLD)
    print(hp_val_from_yaml.ETA_S)
    print(hp_val_from_yaml.STORAGE_T_IN)
    print(hp_val_from_yaml.COLD_STORAGE_T_IN)
    print(hp_val_from_yaml.STORAGE_T_OUT)
    print(hp_val_from_yaml.COLD_STORAGE_T_OUT)
    print(hp_val_from_yaml.STORAGE_P_IN)
    print(hp_val_from_yaml.COLD_STORAGE_P_IN)
    print(hp_val_from_yaml.Q_DOT_MIN)
    print(hp_val_from_yaml.D_T_SUPER)
    print(hp_val_from_yaml.D_T_MIN)



# -*- coding: utf-8 -*-
"""
Created on Fri Jul 26 13:26:46 2024

@author: atakan
"""

import json
import numpy as np
import yaml
import os

class Serializable:
    default_path = ''
    default_filename = 'default.json'
    
    @property
    def all(self):
        return {key: value for key, value in self.__dict__.items() if key != 'all'}

    def _is_serializable(self, value):
        try:
            json.dumps(value)
            return True
        except (TypeError, OverflowError):
            return False

    def _to_serializable(self, value):
        if isinstance(value, np.ndarray):
            return value.tolist()
        elif isinstance(value, (np.float64, np.int64)):
            return value.item()
        elif hasattr(value, 'to_dict'):
            return value.to_dict()
        elif isinstance(value, dict):
            return {k: self._to_serializable(v) for k, v in value.items() if self._is_serializable(v)}
        elif isinstance(value, list):
            return [self._to_serializable(item) for item in value if self._is_serializable(item)]
        elif isinstance(value, (int, float, str, bool)):
            return value
        else:
            return str(value)

    def to_dict(self):
        result = {}
        for key, value in self.__dict__.items():
            result[key] = self._to_serializable(value)
        return result

    @classmethod
    def from_dict(cls, data):
        instance = cls.__new__(cls)
        for key, value in data.items():
            if isinstance(value, list) and len(value) > 0 and isinstance(value[0], (int, float)):
                setattr(instance, key, np.array(value))
            elif isinstance(value, dict):
                if 'optimize' in value or 'setting' in value:  # Hier kannst du die spezifischen Schlüssel überprüfen
                    setattr(instance, key, value)
                else:
                    setattr(instance, key, cls.from_dict(value))
            else:
                setattr(instance, key, value)
        return instance

    def save_to_file(self, file_path=None):
        if file_path is None:
            file_path = os.path.join(self.default_path, self.default_filename)
        try:
            data = self.to_dict()
            if file_path.endswith('.json'):
                with open(file_path, 'w') as file:
                    json.dump(data, file, indent=4)
            elif file_path.endswith('.yaml') or file_path.endswith('.yml'):
                with open(file_path, 'w') as file:
                    yaml.dump(data, file, default_flow_style=False)
            else:
                raise ValueError("Unsupported file format. Use .json or .yaml/.yml")
        except Exception as e:
            print(f"Error saving to file: {e}")

    @classmethod
    def load_from_file(cls, file_path=None):
        if file_path is None:
            file_path = os.path.join(cls.default_path, cls.default_filename)
        try:
            if file_path.endswith('.json'):
                with open(file_path, 'r') as file:
                    data = json.load(file)
            elif file_path.endswith('.yaml') or file_path.endswith('.yml'):
                with open(file_path, 'r') as file:
                    data = yaml.load(file, Loader=yaml.SafeLoader)
            else:
                raise ValueError("Unsupported file format. Use .json or .yaml/.yml")

            return cls.from_dict(data)
        except Exception as e:
            print(f"Error loading from file: {e}")
            return None


# Beisp
class ExampleClass(Serializable):
    default_path = r'C:\Users\atakan\sciebo\results'
    default_filename = 'example_class.json'

    def __init__(self):
        self.D_T_MIN = 4.0
        self.D_T_SUPER = 5
        self.ETA_S_C = 0.7
        self.FLUID_COLD = 'Methanol'
        self.FLUID_STORAGE = 'Water'
        self.FLUID_WORKING = 'Propane * Butane * Pentane * Hexane'
        self.P_IN_STORAGE_COLD = 500000.0
        self.P_IN_STORAGE_HOT = 500000.0
        self.P_WORKING = {'optimize': 'p and x',
                          'setting': 'fixed-p',
                          'p_low': np.float64(88596.5138086614),
                          'p_high': np.float64(1468265.9209155128)}
        self.Q_DOT_MIN = 1000.0
        self.T_IN_STORAGE_COLD = 288.15
        self.T_IN_STORAGE_HOT = 288.15
        self.T_OUT_STORAGE_COLD = 250.15
        self.T_OUT_STORAGE_HOT = 363.0
        self.fluids_all = [['WORKING', 'Propane * Butane * Pentane * Hexane', np.array([0.675, 0.0515, 0.2735])],
                           ['STORAGE_HOT', 'Water', [1.0]],
                           ['STORAGE_COLD', 'Methanol', [1.0]]]
        self.p_low = np.float64(88596.5138086614)
        self.p_high = np.float64(1468265.9209155128)
        self.h_h_out_w = np.float64(143340.64571347312)
        self.h_l_out_w = np.float64(498179.4484371684)
        self.h_h_out_sec = np.float64(376741.38712295156)
        self.h_l_out_cold = np.float64(-220917.46629775176)
        self.T_hh = 363.0

# Erstellen einer Instanz und Speichern in einer Datei
example = ExampleClass()
example.save_to_file()

# Laden aus der Datei
loaded_example = ExampleClass.load_from_file()
print(loaded_example.to_dict())



    